
from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from .http.response import Response


class ArgExcMixin(Exception):
    """
    Mixin exception class where the `arg` argument is used
    as the exception message.
    """
    def __init__(self, arg: object = None) -> None:
        super().__init__()
        self.arg: object = arg

    def __str__(self) -> str:
        if self.arg is None:
            return self.get_default_message()
        return str(self.arg)

    def get_default_message(self) -> str:
        """Get a default message for this exception type."""
        return ''

class ArgExc(ArgExcMixin):
    pass


class OperationException(ArgExc):
    """
    A class of exceptions for when the client needs to raise
    an artificial exception.
    """

class NoResultException(OperationException):
    """Raised when a requested target does not exist."""

class RejectedResultException(OperationException):
    """Raised when a returned value does not fulfil some invariant."""

class UnexpectedResultException(OperationException):
    """Raised when a certain result was not expected."""


class Throwaway(ArgExc):
    """These exceptions are not intended to be caught."""

class UserAgentRequired(Throwaway):
    """Raised when the client detects that the Reddit API wants you to set a user agent."""


def raise_for_non_json_response(resp: Response) -> None:
    """Raise exceptions for a HTTP response from Reddit that does not contain JSON.

    This function assumes the given response object does not contain JSON.
    """
    data = resp.data
    is_html_content = resp.headers.get('Content-Type', '').startswith('text/html')

    if is_html_content:
        if b"user agent required" in data:
            raise UserAgentRequired('the Reddit API wants you to set a user agent')

    try:
        resp.ensure_successful_status()
    except http.exceptions.StatusCodeException as e:
        if is_html_content:
            msg = None
            if b"Our CDN was unable to reach our servers" in data:
                msg = 'HTML, "Our CDN was unable to reach our servers"'
            if b"title>reddit.com: page not found</title" in data:
                msg = 'HTML, page not found'
            e.arg = msg
        raise


class APIError(OperationException):
    """A formal API error."""

class RedditError(APIError):
    """A Reddit error.

    Errors from Reddit's API typically consist of three pieces of information:
    an error label, an explanation, and the name of a related parameter field.
    """

    __match_args__ = ('label',)

    def __init__(self,
        arg: object = None,
        *,
        label: str,
        explanation: str,
        field: str,
    ) -> None:
        super().__init__(arg)
        self.label: str = label
        ("""
            A label for the error. E.g., `USER_REQUIRED`, `INVALID_OPTION`, `SUBREDDIT_NOEXIST`.
            In rare cases this label may not always be in uppercase. It can even contain spaces.
            """)
        self.explanation: str = explanation
        ("""
            A description for the error.
            """)
        self.field: str = field
        ("""
            The name of the parameter relevant to the error, if applicable.
            """)

    def get_default_message(self) -> str:
        la = self.label
        xp = self.explanation
        fd = self.field
        if la:
            if xp:
                if fd:
                    return f'{la}: {xp!r} -> {fd}'
                return f'{la}: {xp!r}'
            if fd:
                return f'{la} -> {fd}'
            return la
        return ''

VanillaProviderAPIError = RedditError


def raise_for_reddit_error(json_data: Any) -> None:
    """Examine JSON data returned from the API and raise exceptions if
    any API errors were detected.

    This function is the default `snub` parameter value for the
    `Client.request` method.
    """
    if not isinstance(json_data, Mapping):
        return

    try:
        if (
            isinstance(label := json_data.get('reason'), str)
            and isinstance(explanation := json_data.get('explanation'), str)
            and isinstance(field := next(iter(json_data.get('fields', [])), None), str)
        ):
            raise RedditError(label=label, explanation=explanation, field=field)
        elif (
            isinstance(label := json_data.get('reason'), str)
            and json_data.get('explanation') is None
            and isinstance(field := next(iter(json_data.get('fields', [])), None), str)
        ):
            raise RedditError(label=label, explanation='', field=field)
        elif (
            isinstance(reason := json_data.get('reason'), str)
            and json_data.get('explanation') is None
            and list(json_data.get('fields', [])) == [None]
        ):
            raise RedditError(label=reason, explanation='', field='')
        elif (
            isinstance(label := json_data.get('reason'), str)
            and isinstance(explanation := json_data.get('explanation'), str)
        ):
            raise RedditError(label=label, explanation=explanation, field='')
        elif json_data.keys() >= {'error', 'message'} and isinstance(reason := json_data.get('reason'), str):
            raise RedditError(label=reason, explanation='', field='')
        elif json_data.keys() >= {'error', 'message'}:
            # No useful information. This will be treated as a StatusCodeException.
            return
        elif json_data.keys() >= {'message'} and isinstance(reason := json_data.get('reason'), str):
            raise RedditError(label=reason, explanation='', field='')
        elif (
            (error_record := next(iter(json_data.get('json', {}).get('errors', [])), [None, None, None]))
            and isinstance(label := error_record[0], str)
            and isinstance(explanation := error_record[1], str)
            and isinstance(field := error_record[2], str)
        ):
            raise RedditError(label=label, explanation=explanation, field=field)
        elif (
            (error_record := next(iter(json_data.get('json', {}).get('errors', [])), [None, None, None]))
            and isinstance(label := error_record[0], str)
            and isinstance(explanation := error_record[1], str)
        ):
            raise RedditError(label=label, explanation=explanation, field='')
        elif (
            isinstance(label := next(iter(json_data.get('errors', [])), None), str)
            and isinstance(explanation := next(iter(json_data.get('errors_values', [])), None), str)
        ):
            raise RedditError(label=label, explanation=explanation, field='')
    except AttributeError:
        pass


from . import http  # Avoid cyclic import




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from ..models.submission_collection_ASYNC import SubmissionCollectionInfo, SubmissionCollection

def load_submission_collection_info(d: Mapping[str, Any], client: Client) -> SubmissionCollectionInfo:
    return SubmissionCollectionInfo(d, client)

def load_submission_collection(d: Mapping[str, Any], client: Client) -> SubmissionCollection:
    return SubmissionCollection(d, client)




from __future__ import annotations
from typing import Any, Sequence

from ..models.report import ModReport, UserReport

def load_mod_report(arr: Sequence[Any]) -> ModReport:
    reason, agent_name = arr
    return ModReport(
        reason=reason,
        agent_name=agent_name,
    )

def load_user_report(arr: Sequence[Any]) -> UserReport:
    reason, count, snoozed, can_snooze = arr
    return UserReport(
        reason=reason,
        count=count,
        snoozed=snoozed,
        can_snooze=can_snooze,
    )




from __future__ import annotations
from typing import Mapping, Any, Optional

from ..models.widget.SYNC import (
    Widget,
    TextAreaWidget,
    ButtonWidget,
    ImageWidget,
    CommunityListWidget,
    CalendarWidget,
    PostFlairWidget,
    CustomCSSWidget,
    CommunityDetailsWidget,
    ModeratorListWidget,
    RulesWidget,
    MenuBar,
)
from ..models.widget.image_size_named_tuple import ImageSize
from ..models.widget.image import ImageWidgetItem
from ..models.widget.button import (
    HoverState,
    TextHoverState,
    ImageHoverState,
    Button,
    TextButton,
    ImageButton,
)
from ..models.widget.community_list import CommunityListWidgetItem
from ..models.widget.calendar import CalendarWidgetConfiguration
from ..models.widget.post_flair import PostFlairWidgetItem
from ..models.widget.custom_css import CustomCSSWidgetImageInfo
from ..models.widget.moderator_list import ModeratorInfo
from ..models.widget.rules import Rule
from ..models.widget.menu_bar import Tab, LinkTab, SubmenuItem, SubmenuTab


def load_widget(d: Mapping[str, Any]) -> Widget:
    kind = d['kind']
    if kind == 'textarea':
        return load_text_area_widget(d)
    elif kind == 'button':
        return load_button_widget(d)
    elif kind == 'image':
        return load_image_widget(d)
    elif kind == 'community-list':
        return load_community_list_widget(d)
    elif kind == 'calendar':
        return load_calendar_widget(d)
    elif kind == 'post-flair':
        return load_post_flair_widget(d)
    elif kind == 'custom':
        return load_custom_css_widget(d)
    elif kind == 'id-card':
        return load_community_details_widget(d)
    elif kind == 'moderators':
        return load_moderator_list_widget(d)
    elif kind == 'subreddit-rules':
        return load_rules_widget(d)
    raise Exception

def load_text_area_widget(d: Mapping[str, Any]) -> TextAreaWidget:
    styles = d['styles']
    return TextAreaWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        text=d['text'],
        text_html=d['textHtml'],
    )

def load_button_widget(d: Mapping[str, Any]) -> ButtonWidget:
    styles = d['styles']
    buttons: list[Button] = []
    for m in d['buttons']:
        hover_state: Optional[HoverState] = None
        hs = m['hoverState']
        if hs is not None:
            if hs['kind'] == 'text':
                hover_state = TextHoverState(
                    label=hs['text'],
                    text_color=hs.get('textColor', ''),
                    fill_color=hs.get('fillColor', ''),
                    stroke_color=hs['color'],
                )
            else:
                hover_state = ImageHoverState(
                    image_url=hs['text'],
                    image_size=ImageSize(hs['width'], hs['height']),
                )

        btn: Button
        if m['kind'] == 'text':
            btn = TextButton(
                label=m['text'],
                link=m['url'],
                hover_state=hover_state,
                text_color=m.get('textColor', ''),
                fill_color=m.get('fillColor', ''),
                stroke_color=m['color'],
            )
            buttons.append(btn)
        else:
            btn = ImageButton(
                label=m['text'],
                link=m['linkUrl'],
                hover_state=hover_state,
                image_url=m['url'],
                image_size=ImageSize(m['width'], m['height']),
            )
            buttons.append(btn)

    return ButtonWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        description=d['description'],
        description_html=d['descriptionHtml'],
        buttons=buttons,
    )

def load_image_widget(d: Mapping[str, Any]) -> ImageWidget:
    styles = d['styles']
    image_items = [
        ImageWidgetItem(
            url=m['url'],
            size=ImageSize(m['width'], m['height']),
            link=m['linkUrl'] or '',
        )
        for m in d['data']
    ]
    return ImageWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        items=image_items,
    )

def load_community_list_widget(d: Mapping[str, Any]) -> CommunityListWidget:
    styles = d['styles']
    subr_items = [
        CommunityListWidgetItem(
            name=m['name'],
            subscribers=m['subscribers'],
            icon_img=m['iconUrl'],
            community_icon=m['communityIcon'],
            primary_color=m['primaryColor'],
            nsfw=m['isNSFW'],
        )
        for m in d['data']
    ]
    return CommunityListWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        items=subr_items,
    )

def load_calendar_widget(d: Mapping[str, Any]) -> CalendarWidget:
    styles = d['styles']
    m = d['configuration']
    configuration = CalendarWidgetConfiguration(
        num_events=m['numEvents'],
        show_title=m['showTitle'],
        show_description=m['showDescription'],
        show_location=m['showLocation'],
        show_date=m['showDate'],
        show_time=m['showTime'],
    )
    return CalendarWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        google_calendar_id=d['googleCalendarId'],
        requires_sync=d['requiresSync'],
        configuration=configuration,
    )

def load_post_flair_widget(d: Mapping[str, Any]) -> PostFlairWidget:
    styles = d['styles']
    templates = [
        PostFlairWidgetItem(
            d=m,
            uuid=m['templateId'],
            text_mode=m['type'],
            text=m['text'],
            bg_color=m['backgroundColor'],
            fg_color_scheme=m['textColor'],
        )
        for m in d['templates']
    ]
    return PostFlairWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        display=d['display'],
        order=d['order'],
        templates=templates,
    )

def load_custom_css_widget(d: Mapping[str, Any]) -> CustomCSSWidget:
    styles = d['styles']
    image_data = [
        CustomCSSWidgetImageInfo(
            url=m['url'],
            width=m['width'],
            height=m['height'],
            name=m['name'],
        )
        for m in d['imageData']
    ]
    return CustomCSSWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        text=d['text'],
        css=d['css'],
        height=d['height'],
        image_data=image_data,
    )

def load_community_details_widget(d: Mapping[str, Any]) -> CommunityDetailsWidget:
    styles = d['styles']
    return CommunityDetailsWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        public_description=d.get('description', ''),
        subscriber_text=d['subscribersText'],
        viewing_text=d['currentlyViewingText'],
        subscriber_count=d.get('subscribersCount', ''),
        viewing_count=d.get('currentlyViewingCount', ''),
    )

def load_moderator_list_widget(d: Mapping[str, Any]) -> ModeratorListWidget:
    styles = d['styles']
    mods = [
        ModeratorInfo(
            name=m['name'],
            flair=ModeratorInfo.Flair(
                text_mode=m['authorFlairType'],
                text=m['authorFlairText'] or '',
                bg_color=m['authorFlairBackgroundColor'],
                fg_color_scheme=m['authorFlairTextColor'],
                has_had_flair_assigned_before_in_subreddit=m['authorFlairText'] is not None,
            ),
        )
        for m in d.get('mods', ())
    ]
    return ModeratorListWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title='',
        mod_num=d.get('totalMods', 0),
        mods=mods,
    )

def load_rules_widget(d: Mapping[str, Any]) -> RulesWidget:
    styles = d['styles']
    rules = [
        Rule(
            description=m['description'],
            description_html=m['descriptionHtml'],
            short_name=m['shortName'],
            violation_reason=m['violationReason'],
            created_ut=m['createdUtc'],
        )
        for m in d.get('data', ())
    ]
    return RulesWidget(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        header_color=styles['headerColor'] or '',
        background_color=styles['backgroundColor'] or '',
        title=d['shortName'],
        display=d['display'],
        rules=rules,
    )


def load_menu_bar(d: Mapping[str, Any]) -> MenuBar:
    tabs: list[Tab] = []
    tab: Tab
    for m in d['data']:
        if 'children' in m:
            tab = SubmenuTab(
                label=m['text'],
                items=[
                    SubmenuItem(
                        label=mi['text'],
                        link=mi['url'],
                    )
                    for mi in m['children']
                ],
            )
            tabs.append(tab)
        else:
            tab = LinkTab(
                label=m['text'],
                link=m['url'],
            )
            tabs.append(tab)

    return MenuBar(
        d=d,
        idt=d['id'],
        kind=d['kind'],
        show_wiki=d['showWiki'],
        tabs=tabs,
    )




from __future__ import annotations

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class ModeratorInfo:
    @dataclass(repr=False, eq=False)
    class Flair:
        text_mode: str
        text: str
        bg_color: str
        fg_color_scheme: str
        has_had_flair_assigned_before_in_subreddit: bool

    name: str
    flair: Flair




from typing import Any, Mapping, Optional

from datetime import datetime
from dataclasses import dataclass

from .datamemento import DatamementoDataclassesMixin

@dataclass(repr=False, eq=False)
class ModerationNote(DatamementoDataclassesMixin):
    #_: KW_ONLY
    d: Mapping[str, Any]
    uuid: str
    unixtime: int
    ("""
        Unix timestamp of when the mod note entry was made.
        """)
    datetime: datetime
    ("""
        Datetime object of when the mod note entry was made.
        """)
    label: str
    subreddit_id: int
    ("""
        ID of the subreddit in which this note belongs.
        """)
    subreddit: str
    ("""
        Name of the subreddit in which this note belongs.
        """)
    agent_id: int
    ("""
        User ID of the moderator associated with this note.
        """)
    agent: str
    ("""
        Name of the moderator associated with this note.
        """)
    target_id: int
    ("""
        ID of the user who this note applies to.
        """)
    target: str
    ("""
        Name of the user who this note applies to.
        """)

@dataclass(repr=False, eq=False)
class ModerationActionNote(ModerationNote):
    label: str
    ("""
        Values: `APPROVAL`, `REMOVAL`, `BAN`, `MUTE`,
        `INVITE`, `SPAM`, `CONTENT_CHANGE`, `MOD_ACTION`.
        """)
    action: str
    ("""
        Name of the mod log action. In lowercase.
        """)

@dataclass(repr=False, eq=False)
class ModerationUserNote(ModerationNote):
    label: str
    ("""
        Values: empty string, `ABUSE_WARNING`, `SPAM_WARNING`, `SPAM_WATCH`,
        `SOLID_CONTRIBUTOR`, `HELPFUL_USER`, `BOT_BAN`, `PERMA_BAN`, `BAN`.

        Value empty string if no label was assigned to this note.
        """)
    note: str
    ("""
        Content of the user note.
        """)
    has_anchor: bool
    ("""
        Whether this note is associated with a submission or comment.

        If true then either :attr:`anchor_submission_id` or :attr:`anchor_comment_id` will be non-`None`.
        """)
    anchor_submission_id: Optional[int]
    ("""
        The submission ID associated with this user note.
        """)
    anchor_comment_id: Optional[int]
    ("""
        The comment ID associated with this user note.
        """)




from __future__ import annotations
from typing import Mapping, ClassVar

from http import HTTPStatus

from ..exceptions import ArgExcMixin

class ArgExc(ArgExcMixin):
    pass


class TransportError(ArgExc):
    """A catch-all exception for HTTP transport library errors."""

class TimeoutException(ArgExc):
    """A timeout exception class to wrap timeout errors from HTTP transport libraries."""

class StatusCodeException(ArgExc):
    """An exception class representing HTTP status codes as errors."""

    __match_args__ = ('status_code',)

    STATUS_CODE: ClassVar[int] = 0

    def __init__(self, arg: object = None, *, status_code: int) -> None:
        super().__init__(arg)
        self.status_code: int = status_code
        ("")

    def get_default_message(self) -> str:
        sts = self.status_code
        try:
            return f"{sts} {HTTPStatus(sts).phrase}"
        except ValueError:
            return f"{sts}"

class StatusCodeExceptionTypes:
    class InformationalStatusCodeException(StatusCodeException):
        STATUS_CODE: ClassVar[int] = -100
    class SuccessfulStatusCodeException(StatusCodeException):
        STATUS_CODE: ClassVar[int] = -200
    class RedirectionStatusCodeException(StatusCodeException):
        STATUS_CODE: ClassVar[int] = -300
    class ClientErrorStatusCodeException(StatusCodeException):
        STATUS_CODE: ClassVar[int] = -400
    class ServerErrorStatusCodeException(StatusCodeException):
        STATUS_CODE: ClassVar[int] = -500

    class BadRequest(ClientErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 400
    class Unauthorized(ClientErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 401
    class Forbidden(ClientErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 403
    class NotFound(ClientErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 404
    class Conflict(ClientErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 409
    class PayloadTooLarge(ClientErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 413
    class URITooLong(ClientErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 414
    class TooManyRequests(ClientErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 429

    class InternalServerError(ServerErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 500
    class BadGateway(ServerErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 502
    class ServiceUnavailable(ServerErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 503
    class GatewayTimeout(ServerErrorStatusCodeException):
        STATUS_CODE: ClassVar[int] = 504

status_code_exception_class_by_status_code: Mapping[int, type[StatusCodeException]] = {
    cls.STATUS_CODE: cls
    for cls in [
        # Client errors
        StatusCodeExceptionTypes.BadRequest,
        StatusCodeExceptionTypes.Unauthorized,
        StatusCodeExceptionTypes.Forbidden,
        StatusCodeExceptionTypes.NotFound,
        StatusCodeExceptionTypes.Conflict,
        StatusCodeExceptionTypes.PayloadTooLarge,
        StatusCodeExceptionTypes.URITooLong,
        StatusCodeExceptionTypes.TooManyRequests,

        # Server errors
        StatusCodeExceptionTypes.InternalServerError,
        StatusCodeExceptionTypes.BadGateway,
        StatusCodeExceptionTypes.ServiceUnavailable,
        StatusCodeExceptionTypes.GatewayTimeout,
    ]
}

def get_status_code_exception_class_by_status_code(n: int) -> type[StatusCodeException]:
    klass = status_code_exception_class_by_status_code.get(n)
    if klass is None:
        klass = StatusCodeException
        if 100 <= n <= 199:
            klass = StatusCodeExceptionTypes.InformationalStatusCodeException
        elif 200 <= n <= 299:
            klass = StatusCodeExceptionTypes.SuccessfulStatusCodeException
        elif 300 <= n <= 399:
            klass = StatusCodeExceptionTypes.RedirectionStatusCodeException
        elif 400 <= n <= 499:
            klass = StatusCodeExceptionTypes.ClientErrorStatusCodeException
        elif 500 <= n <= 599:
            klass = StatusCodeExceptionTypes.ServerErrorStatusCodeException
    return klass


def is_successful_status(n: int) -> bool:
    """Return true if `200 <= n < 300`."""
    return 200 <= n <= 299

def raise_now(n: int) -> None:
    """Raises a :class:`.StatusCodeException` exception type based on the given HTTP status number."""
    raise get_status_code_exception_class_by_status_code(n)(status_code=n)

def ensure_successful_status(n: int) -> None:
    """Raises a :class:`.StatusCodeException` exception if `is_successful_status(n)` returns false."""
    if not is_successful_status(n):
        raise_now(n)




from __future__ import annotations
import mimetypes

def guess_filename_mimetype(filename: str) -> str:
    y = mimetypes.guess_type(filename, strict=False)[0]
    return y or 'application/octet-stream'




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable, Sequence, Any, Mapping
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.comment_SYNC import LooseComment
    from ...models.subreddit_rules import SubredditRules
    from ...types import JSON_ro

from ...models.subreddit_SYNC import Subreddit, InaccessibleSubreddit
from ...model_loaders.subreddit_SYNC import load_subreddit, load_potentially_inaccessible_subreddit
from ...model_loaders.subreddit_rules import load_subreddit_rules
from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_iterator import CallChunkChainingIterator
from ...iterators.call_chunk import CallChunk
from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.listing.comment_listing_paginator import LooseCommentListingPaginator
from ...pagination.paginators.subreddit_sync1 import SubredditSearchPaginator
from ... import exceptions
from ... import http
from .pull_SYNC import Pull
from .pulls_SYNC import Pulls

class SubredditProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.pull: Pull = Pull(client)
        ("""
            Pull subreddit submissions.
            """)
        self.pulls: Pulls = Pulls(client)
        ("""
            Pull subreddits.
            """)

    def get(self, idn: int) -> Optional[Subreddit]:
        """Get information about a potentially inaccessible subreddit.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :returns:
            * `None` if the subreddit does not exist, or the subreddit is private or banned.
            * :class:`~.models.subreddit_SYNC.Subreddit`
        :rtype: `Optional`\\[:class:`~.models.subreddit_SYNC.Subreddit`]
        """
        v = self.get_potentially_inaccessible(idn)
        if isinstance(v, Subreddit):
            return v
        return None

    def fetch(self, idn: int) -> Subreddit:
        """Fetch information about a potentially inaccessible subreddit.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: :class:`~.models.subreddit_SYNC.Subreddit`

        .. .RAISES

        :raises redditwarp.exceptions.OperationException:
           - The target subreddit is was not found.
           - The target subreddit is private or banned.
        """
        v = self.fetch_potentially_inaccessible(idn)
        if isinstance(v, Subreddit):
            return v
        if isinstance(v, InaccessibleSubreddit):
            raise exceptions.RejectedResultException
        raise Exception

    def get_by_name(self, name: str) -> Optional[Subreddit]:
        """Return information about a subreddit by name.

        Returns `None` when the target subreddit is private or banned, or
        if the specified subreddit name contained invalid characters, or
        you specified the name of a special subreddit:
        `all`, `popular`, `friends`, `mod`.

        .. .PARAMETERS

        :param `str` name:
            Subreddit name.

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.subreddit_SYNC.Subreddit`]
        """
        try:
            root = self._client.request('GET', f'/r/{name}/about')
        except exceptions.RedditError as e:
            if e.label.lower() in ('private', 'banned'):
                return None
            raise
        except http.exceptions.StatusCodeException as e:
            if e.status_code in (302, 404):
                return None
            raise
        if root['kind'] == 'Listing':
            return None
        return load_subreddit(root['data'], self._client)

    def fetch_by_name(self, name: str) -> Subreddit:
        """Return information about a subreddit by name.

        Behaves similarly to :meth:`.get_by_name`.

        .. .PARAMETERS

        :param `str` name:
            Subreddit name.

        .. .RETURNS

        :rtype: :class:`~.models.subreddit_SYNC.Subreddit`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `private`:
                The target subreddit is private.
            + `banned`:
                The target subreddit is banned.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `302`:
               - The subreddit does not exist.

            + `404`:
               - You specified the name of a special subreddit: `all`, `popular`, `friends`, `mod`.
               - The specified subreddit name was too long or contained invalid characters.
        """
        root = self._client.request('GET', f'/r/{name}/about')
        if root['kind'] == 'Listing':
            raise exceptions.NoResultException('target not found')
        return load_subreddit(root['data'], self._client)

    def get_potentially_inaccessible(self, idn: int) -> Optional[object]:
        """Get information about a potentially inaccessible subreddit.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :returns:
            * :class:`~.models.subreddit_SYNC.InaccessibleSubreddit` if the subreddit is private or banned.
            * `None` if the subreddit does not exist.
            * :class:`~.models.subreddit_SYNC.Subreddit`
        :rtype: `Optional`\\[`object`]
        """
        id36 = to_base36(idn)
        full_id36 = 't5_' + id36
        root = self._client.request('GET', '/api/info', params={'id': full_id36})
        if children := root['data']['children']:
            return load_potentially_inaccessible_subreddit(children[0]['data'], self._client)
        return None

    def fetch_potentially_inaccessible(self, idn: int) -> object:
        """Fetch information about a potentially inaccessible subreddit.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :returns:
            * :class:`~.models.subreddit_SYNC.InaccessibleSubreddit` if the subreddit is private or banned.
            * `None` if the subreddit does not exist.
            * :class:`~.models.subreddit_SYNC.Subreddit`
        :rtype: `Optional`\\[`object`]

        .. .RAISES

        :raises redditwarp.exceptions.NoResultException:
            The target subreddit is was not found.
        """
        v = self.get_potentially_inaccessible(idn)
        if v is None:
            raise exceptions.NoResultException('target not found')
        return v

    def get_potentially_inaccessible_by_name(self, name: str) -> Optional[object]:
        """Get information about a potentially inacessible subreddit.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :returns:
            * :class:`~.models.subreddit_SYNC.InaccessibleSubreddit` if the subreddit is private or banned.
            * `None` if the subreddit does not exist.
            * :class:`~.models.subreddit_SYNC.Subreddit`
        :rtype: `Optional`\\[`object`]
        """
        root = self._client.request('GET', '/api/info', params={'sr_name': name})
        if children := root['data']['children']:
            return load_potentially_inaccessible_subreddit(children[0]['data'], self._client)
        return None

    def fetch_potentially_inaccessible_by_name(self, name: str) -> object:
        """Fetch information about a potentially inacessible subreddit.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :returns:
            * :class:`~.models.subreddit_SYNC.InaccessibleSubreddit` if the subreddit is private or banned.
            * `None` if the subreddit does not exist.
            * :class:`~.models.subreddit_SYNC.Subreddit`
        :rtype: `object`

        .. .RAISES

        :raises redditwarp.exceptions.NoResultException:
            The target subreddit is was not found.
        """
        v = self.get_potentially_inaccessible_by_name(name)
        if v is None:
            raise exceptions.NoResultException('target not found')
        return v

    def bulk_fetch_potentially_inaccessible(self, ids: Iterable[int]) -> CallChunkChainingIterator[object]:
        """Bulk fetch information about a potentially inacessible subreddits, by ID.

        Any ID not found will be ignored.

        .. .PARAMETERS

        :param `Iterable[int]` ids:

        .. .RETURNS

        :returns:
            Iterator of:

            * :class:`~.models.subreddit_SYNC.InaccessibleSubreddit` if the subreddit is private or banned.
            * `None` if the subreddit does not exist.
            * :class:`~.models.subreddit_SYNC.Subreddit`
        :rtype: :class:`~.iterators.call_chunk_chaining_iterator.CallChunkChainingIterator`\\[`object`]
        """
        def mass_fetch(ids: Sequence[int]) -> Sequence[object]:
            id36s = map(to_base36, ids)
            full_id36s = map('t5_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            root = self._client.request('GET', '/api/info', params={'id': ids_str})
            return [load_potentially_inaccessible_subreddit(i['data'], self._client) for i in root['data']['children']]

        return CallChunkChainingIterator(CallChunk(mass_fetch, chunk) for chunk in chunked(ids, 100))

    def bulk_fetch_potentially_inaccessible_by_name(self, names: Iterable[str]) -> CallChunkChainingIterator[object]:
        """Bulk fetch information about a potentially inacessible subreddits, by name.

        Any name not found will be ignored.

        .. .PARAMETERS

        :param `Iterable[int]` ids:

        .. .RETURNS

        :returns:
            Iterator of:

            * :class:`~.models.subreddit_SYNC.InaccessibleSubreddit` if the subreddit is private or banned.
            * `None` if the subreddit does not exist.
            * :class:`~.models.subreddit_SYNC.Subreddit`
        :rtype: :class:`~.iterators.call_chunk_chaining_iterator.CallChunkChainingIterator`\\[`object`]
        """
        def mass_fetch(names: Sequence[str]) -> Sequence[object]:
            root = self._client.request('GET', '/api/info', params={'sr_name': ','.join(names)})
            return [load_potentially_inaccessible_subreddit(i['data'], self._client) for i in root['data']['children']]

        return CallChunkChainingIterator(CallChunk(mass_fetch, chunk) for chunk in chunked(names, 100))

    def pull_new_comments(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[LooseCommentListingPaginator, LooseComment]:
        """Pull new comments from a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `Optional[int]` amount:

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`\\[:class:`~.pagination.paginators.listing.comment_listing_paginator.LooseCommentListingPaginator`, :class:`~.models.comment_SYNC.LooseComment`]
        """
        p = LooseCommentListingPaginator(self._client, f'/r/{sr}/comments')
        return ImpartedPaginatorChainingIterator(p, amount)

    def get_settings(self, sr: str) -> Mapping[str, Any]:
        """Get the settings of a subreddit.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `Mapping`\\[`str`, `Any`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `302`:
                The specified subreddit does not exist.
            + `404`:
               - There is no user context.
               - You don't have permission to view this subreddit's settings.
        """
        root = self._client.request('GET', f'/r/{sr}/about/edit')
        if root['kind'] != 'subreddit_settings':
            raise exceptions.NoResultException('target not found')
        return root['data']

    def update_settings(self, idn: int, settings: Mapping[str, JSON_ro]) -> None:
        """Update a subreddit's settings.

        .. .PARAMETERS

        :param `int` idn:
        :param settings:
        :type settings: `Mapping`\\[`str`, :class:`~.types.JSON_ro`]

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `SUBREDDIT_REQUIRED`:
                The specified subreddit does not exist.
            + `MOD_REQUIRED`:
                The current user is not a moderator of the subreddit.
        """
        id36 = to_base36(idn)
        full_id36 = 't5_' + id36
        json: Mapping[str, JSON_ro] = {**settings, 'sr': full_id36}
        self._client.request('PATCH', '/api/v1/subreddit/update_settings', json=json)

    def subscribe_by_id(self, idn: int) -> None:
        """Subscribe to a subreddit.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The specified subreddit could not be accessed.
            + `404`:
                The specified subreddit does not exist.
        """
        id36 = to_base36(idn)
        full_id36 = 't5_' + id36
        self._client.request('POST', '/api/subscribe', data={'action': 'sub', 'sr': full_id36})

    def subscribe_by_name(self, name: str) -> None:
        """Subscribe to a subreddit.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The specified subreddit could not be accessed.
               - The specified subreddit was a special name such as `popular`, `all`, or `random`.

            + `404`:
                The specified subreddit does not exist.
        """
        self._client.request('POST', '/api/subscribe', data={'action': 'sub', 'sr_name': name})

    def unsubscribe_by_id(self, idn: int) -> None:
        """Unsubscribe from a subreddit.

        Behaves similarly to :meth:`.subscribe_by_id`.
        """
        id36 = to_base36(idn)
        full_id36 = 't5_' + id36
        self._client.request('POST', '/api/subscribe', data={'action': 'unsub', 'sr': full_id36})

    def unsubscribe_by_name(self, name: str) -> None:
        """Unsubscribe from a subreddit.

        Behaves similarly to :meth:`.subscribe_by_name`.
        """
        self._client.request('POST', '/api/subscribe', data={'action': 'unsub', 'sr_name': name})

    def get_rules(self, sr: str) -> SubredditRules:
        """Get a subreddit's rules.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: :class:`~.models.subreddit_rules.SubredditRules`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The specified subreddit could not be accessed.
        """
        root = self._client.request('GET', f'/r/{sr}/about/rules')
        if 'rules' not in root:
            raise exceptions.NoResultException('target not found')
        return load_subreddit_rules(root)

    def get_post_requirements(self, sr: str) -> Mapping[str, Any]:
        """Get the post requirement settings of a subreddit.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `Mapping`\\[`str`, `Any`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `SUBREDDIT_NOEXIST`:
                The specified subreddit does not exist.
            + `SUBREDDIT_NO_ACCESS`:
                The specified subreddit is private or banned.
        """
        return self._client.request('GET', f'/api/v1/{sr}/post_requirements')

    def search(self, query: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[SubredditSearchPaginator, Subreddit]:
        """Search subreddits by name or description.

        .. .PARAMETERS

        :param `str` query:
        :param `Optional[int]` amount:

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`\\[:class:`~.pagination.paginators.subreddit_sync1.SubredditSearchPaginator`, :class:`~.models.subreddit_SYNC.Subreddit`]
        """
        if not query:
            raise ValueError('query cannot be empty')
        p = SubredditSearchPaginator(self._client, '/subreddits/search', query)
        return ImpartedPaginatorChainingIterator(p, amount)

    def search_names(self, name: str) -> Sequence[str]:
        """List subreddit names that begin with a query string.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :rtype: `Sequence`\\[`str`]
        """
        root = self._client.request('GET', '/api/search_reddit_names', params={'query': name})
        return root['names']

    def exists(self, name: str) -> bool:
        """Check whether a subreddit name exists.

        An empty string returns true.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :rtype: `bool`
        """
        try:
            self._client.request('GET', '/api/search_reddit_names', params={'query': name, 'exact': '1'})
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return False
            raise
        return True

    def get_similar_subreddits(self, idn: int, n: Optional[int] = None) -> Sequence[Subreddit]:
        """Get a list of similar subreddits.

        .. .PARAMETERS

        :param `int` idn:
        :param `Optional[int]` n:
            The maximum number of entries to return.

            Effective default: 10.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.subreddit_SYNC.Subreddit`]
        """
        params = {'sr_fullnames': 't5_' + to_base36(idn)}
        if n is not None:
            params['max_recs'] = str(n)
        root = self._client.request('GET', '/api/similar_subreddits', params=params)
        return [load_subreddit(d['data'], client=self._client) for d in root['data']['children']]




from __future__ import annotations
from typing import TYPE_CHECKING, MutableMapping, Sequence, Protocol, MutableSequence, Optional, Mapping
if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec

from dataclasses import dataclass

from ...util.imports import load_spec, import_module_from_spec
from ..websocket_SYNC import WebSocket


class ConnectFunctionProtocol(Protocol):
    def __call__(self,
        url: str,
        *,
        subprotocols: Sequence[str] = (),
        headers: Optional[Mapping[str, str]] = None,
        timeout: float = -2,
    ) -> WebSocket: ...

@dataclass
class TransportInfo:
    adaptor_module_name: str
    name: str
    version: str
    connect: ConnectFunctionProtocol


def load_transport() -> TransportInfo:
    if not transport_info_registry:
        for module_spec in transport_module_spec_list:
            try:
                import_module_from_spec(module_spec)
            except ImportError:
                continue
            break
        else:
            raise ModuleNotFoundError('A websocket transport library needs to be installed.')

    return next(iter(transport_info_registry.values()))

def connect(
    url: str,
    *,
    subprotocols: Sequence[str] = (),
    headers: Optional[Mapping[str, str]] = None,
    timeout: float = -2,
) -> WebSocket:
    connect = load_transport().connect
    return connect(
        url,
        subprotocols=subprotocols,
        headers=headers,
        timeout=timeout,
    )

def register(
    adaptor_module_name: str,
    name: str,
    version: str,
    connect: ConnectFunctionProtocol,
) -> None:
    info = TransportInfo(
        adaptor_module_name=adaptor_module_name,
        name=name,
        version=version,
        connect=connect,
    )
    transport_info_registry[adaptor_module_name] = info

transport_module_spec_list: MutableSequence[ModuleSpec] = [
    load_spec('.carriers.websocket', __package__),
]
transport_info_registry: MutableMapping[str, TransportInfo] = {}




from __future__ import annotations
from typing import TypeVar, MutableMapping, Mapping, MutableSet, Optional, Iterator, IO, Any

from pprint import PrettyPrinter

K = TypeVar('K')
V = TypeVar('V')

class LockedKeysMappingProxy(MutableMapping[K, V]):
    def __init__(self,
        mapping: MutableMapping[K, V],
        locked_keys: Optional[MutableSet[K]] = None,
    ) -> None:
        self._store = mapping
        if locked_keys is None:
            locked_keys = set()
        self.locked_keys: MutableSet[K] = locked_keys
        ("")

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._store})'

    def __contains__(self, item: object) -> bool:
        return item in self._store

    def __iter__(self) -> Iterator[K]:
        return iter(self._store)

    def __len__(self) -> int:
        return len(self._store)

    def __getitem__(self, key: K) -> V:
        return self._store[key]

    def __setitem__(self, key: K, value: V) -> None:
        if key in self.locked_keys:
            return
        self._store[key] = value

    def __delitem__(self, key: K) -> None:
        if key in self.locked_keys:
            return
        del self._store[key]

    @staticmethod
    def _pprint(
        printer: PrettyPrinter,
        obj: Mapping[K, V],
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Mapping[int, Any],
        level: int,
    ) -> None:
        cls_name = obj.__class__.__name__
        stream.write(cls_name + '(')
        printer._format(  # type: ignore[attr-defined]
            dict(obj),
            stream,
            indent + len(cls_name) + 1,
            allowance + 1,
            context,
            level,
        )
        stream.write(')')

    if isinstance(getattr(PrettyPrinter, '_dispatch', None), dict):
        PrettyPrinter._dispatch[__repr__] = _pprint.__func__  # type: ignore[attr-defined]




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, final
if TYPE_CHECKING:
    from typing import Optional, Mapping, Union
    from types import TracebackType
    from .requisition import Requisition
    from .response import Response
    from .types import RequestFiles
    from .handler_SYNC import Handler
    from .exchange import Exchange
    from ..types import JSON_ro
    from .payload import Payload

from urllib.parse import urljoin

from .requisition import make_requisition
from .send_params import SendParams
from .invoker_SYNC import Invoker


DEFAULT_TIMEOUT: float = 100.


class HTTPClient:
    """A class for sending HTTP requests.

    The purpose of the HTTPClient is to be as useful as possible.
    """

    _TSelf = TypeVar('_TSelf', bound='HTTPClient')

    @staticmethod
    def make_requisition(
        verb: str,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: JSON_ro = None,
        files: Optional[RequestFiles] = None,
        payload: Optional[Payload] = None,
    ) -> Requisition:
        return make_requisition(
            verb,
            url,
            params=params,
            headers=headers,
            data=data,
            json=json,
            files=files,
            payload=payload,
        )

    def __init__(self, handler: Handler) -> None:
        self._invoker: Invoker = Invoker(handler)
        self.base_url: str = ''
        ("")
        self.timeout: float = DEFAULT_TIMEOUT
        ("")
        self.follow_redirects: Optional[bool] = False
        ("")

    def __enter__(self: _TSelf) -> _TSelf:
        return self

    def __exit__(self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.close()
        return None

    @final
    def request(self,
        verb: str,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: JSON_ro = None,
        files: Optional[RequestFiles] = None,
        payload: Optional[Payload] = None,
        timeout: float = -2,
        follow_redirects: Optional[bool] = None,
    ) -> Response:
        xchg = self.inquire(verb, url, params=params, headers=headers,
                data=data, json=json, files=files, payload=payload,
                timeout=timeout, follow_redirects=follow_redirects)
        return xchg.response

    @final
    def inquire(self,
        verb: str,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: JSON_ro = None,
        files: Optional[RequestFiles] = None,
        payload: Optional[Payload] = None,
        timeout: float = -2,
        follow_redirects: Optional[bool] = None,
    ) -> Exchange:
        reqi = self.make_requisition(verb, url, params=params, headers=headers,
                data=data, json=json, files=files, payload=payload)
        return self.submit(reqi, timeout=timeout, follow_redirects=follow_redirects)

    @final
    def submit(self,
        reqi: Requisition,
        *,
        timeout: float = -2,
        follow_redirects: Optional[bool] = None,
    ) -> Exchange:
        p = SendParams(
            reqi,
            timeout=timeout,
            follow_redirects=follow_redirects,
        )
        return self._send(p)

    def _send(self, p: SendParams) -> Exchange:
        reqi = p.requisition
        reqi.url = urljoin(self.base_url, reqi.url)
        if p.timeout == -2:
            p.timeout = self.timeout
        if p.follow_redirects is None:
            p.follow_redirects = self.follow_redirects
        return self._invoker.send(p)

    def close(self) -> None:
        self._invoker.close()



"""An ordered set implementation."""

from __future__ import annotations
from typing import TypeVar, Iterator, Iterable, MutableSet

from itertools import islice

T = TypeVar('T')

class OrderedSet(MutableSet[T]):
    """An ordered set.

    An ordered set is a set that remembers insertion order.
    """

    def __init__(self, it: Iterable[T]) -> None:
        store: dict[T, None] = {}
        self._store = store
        for i in it:
            store[i] = None

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({list(self)})'

    def __len__(self) -> int:
        return len(self._store)

    def __contains__(self, item: object) -> bool:
        return item in self._store

    def __iter__(self) -> Iterator[T]:
        return iter(self._store)

    def add(self, value: T) -> None:
        self._store[value] = None

    def discard(self, value: T) -> None:
        self._store.pop(value, None)


class BoundedSet(OrderedSet[T]):
    """An ordered set with a maximum capacity.

    When the maximum size is reached and a new item is added, the oldest item is
    silently evicted first.
    """

    @property
    def capacity(self) -> int:
        return self._capacity

    def __init__(self, it: Iterable[T], capacity: int) -> None:
        if capacity < 0:
            raise ValueError('capacity must be non-negative')
        super().__init__(it)
        self._capacity = capacity
        store = self._store
        keys = list(islice(store, max(len(store) - capacity, 0)))
        for key in keys:
            del store[key]

    def __repr__(self) -> str:
        capacity = self._capacity
        return f'{self.__class__.__name__}({list(self)}, {capacity=})'

    # https://mobile.twitter.com/raymondh/status/1150079085114093568
    def _from_iterable(self, it: Iterable[T]) -> BoundedSet[T]:
        return type(self)(it, capacity=self._capacity)

    def add(self, value: T) -> None:
        super().add(value)
        if len(self) > self._capacity:
            del self._store[next(iter(self._store))]




from __future__ import annotations
from typing import Mapping, Any, Optional, Sequence, TypeVar, Callable, final

from datetime import datetime, timezone

from ..core.const import AUTHORIZATION_BASE_URL
from .datamemento import DatamementoBase
from .report import ModReport, UserReport
from ..model_loaders.report import load_mod_report, load_user_report
from dataclasses import dataclass

class Submission(DatamementoBase):
    class Me:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.saved: bool = d['saved']
            ("""
                Whether the current user has saved the post.

                Value false if there is no user context.
                """)
            self.reply_notifications: bool = d['send_replies']
            ("""
                Whether an inbox message will be sent to you when the submission receives a new top-level comment.

                Value true if you are not the author of the submission.

                Value true if there is no user context.
                """)
            self.voted: int = {False: -1, None: 0, True: 1}[d['likes']]
            ("""
                Value `1` if upvoted, `0` if not voted on, `-1` if downvoted.
                """)
            self.hidden: bool = d['hidden']
            ("""
                Whether the current user has hidden the post.

                Value false if there is not user context.
                """)
            self.is_following_event: bool = d.get('is_followed', False)
            ("""
                True if the submisison has event details and is being followed by the current user.

                Value false if there is no user context.
                """)

    class Author:
        class AuthorFlair:
            def __init__(self, d: Mapping[str, Any]) -> None:
                flair_text: Optional[str] = d['author_flair_text']
                flair_css_class: Optional[str] = d['author_flair_css_class']

                self.template_uuid: Optional[str] = x if (x := d['author_flair_template_id']) else None
                ("""
                    The author's flair template UUID.

                    Value is null if no flair template is being used.
                    """)
                self.text_mode: str = d['author_flair_type']
                ("""
                    Either `text` or `richtext`.
                    """)
                self.text: str = flair_text or ''
                ("""
                    The author's flair text.

                    Check if the value is an empty string to tell if a flair is being used.
                    """)
                self.css_class: str = flair_css_class or ''
                ("""
                    The author's flair CSS class.

                    When a flair template is being used, the value of this field will be
                    that of the CSS class designated by the template. If the flair template
                    does not specify a CSS class then the value will be an empty string.
                    """)
                self.bg_color: str = '' if (x := d['author_flair_background_color']) == 'transparent' else x
                ("""
                    A background color hex string. E.g., `#46d160`.

                    If a flair template is not being used then the value is an empty string.

                    If a flair template is being used and the background color is unset then
                    the value is an empty string.
                    """)
                self.fg_color_scheme: str = d['author_flair_text_color'] or ''
                ("""
                    Either `dark` or `light`, or empty string.

                    Value is empty string if a flair template is not being used.
                    """)

                self.has_had_flair_assigned_before_in_subreddit: bool = flair_text is not None
                ("""
                    Because of quirks in the API, we can tell if the user has ever had a flair assigned before in the subreddit.
                    """)
                self.has_had_flair_css_class_assigned_before_in_subreddit_when_no_flair_template_assigned: bool = flair_css_class is not None
                ("""
                    Because of quirks in the API, we can tell if the user has ever had a flair CSS class
                    assigned before in the subreddit. This only works if a flair template is not being used.
                    """)

        def __init__(self, d: Mapping[str, Any]) -> None:
            self.name: str = d['author']
            ("")
            self.id36: str = d['author_fullname'].split('_', 1)[-1]
            ("")
            self.id: int = int(self.id36, 36)
            ("")
            self.has_premium: bool = d['author_premium']
            ("")
            self.flair: Submission.Author.AuthorFlair = self.AuthorFlair(d)
            ("")

    class Subreddit:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.id36: str = d['subreddit_id'].split('_', 1)[-1]
            ("")
            self.id: int = int(self.id36, 36)
            ("")
            self.name: str = d['subreddit']
            ("")
            self.openness: str = d['subreddit_type']
            ("""
                Either: `public`, `private`, `restricted`, `archived`,
                `employees_only`, `gold_only`, `gold_restricted`, or `user`.
                """)
            self.quarantined: bool = d['quarantine']
            ("")
            self.subscriber_count: int = d['subreddit_subscribers']
            ("")

    class Moderation:
        class Approved:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.by: str = d['approved_by']
                ("""
                    Name of the moderator who approved this comment.
                    """)
                self.ut: int = d['approved_at_utc']
                ("")
                self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)
                ("")

        class Removed:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.by: str = d['banned_by']
                ("""
                    Name of the moderator who removed this comment.
                    """)
                self.ut: int = d['banned_at_utc']
                ("")
                self.at: datetime = datetime.fromtimestamp(self.ut, timezone.utc)
                ("")

        class Reports:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.ignoring: bool = d['ignore_reports']
                ("")
                self.num_reports: int = d['num_reports']
                ("")
                self.mod_reports: Sequence[ModReport] = [load_mod_report(m) for m in d['mod_reports']]
                ("")
                self.user_reports: Sequence[UserReport] = [load_user_report(m) for m in d['user_reports']]
                ("")

        class RemovalReason:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.by: str = d['mod_reason_by'] or ''
                ("")
                self.title: str = d['mod_reason_title'] or ''
                ("")
                self.note: str = d['mod_note'] or ''
                ("")

        def __init__(self, d: Mapping[str, Any]) -> None:
            self.spam: bool = d['spam']
            ("")

            self.approved: Optional[Submission.Moderation.Approved] = None
            ("")
            if d['approved_by']:
                self.approved = self.Approved(d)

            self.removed: Optional[Submission.Moderation.Removed] = None
            ("")
            if d['banned_by']:
                self.removed = self.Removed(d)

            self.reports: Submission.Moderation.Reports = self.Reports(d)
            ("")

            self.removal_reason: Optional[Submission.Moderation.RemovalReason] = None
            ("")
            if d['mod_reason_by']:
                self.removal_reason = self.RemovalReason(d)

    class Event:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.start_ut: int = int(d['event_start'])
            ("")
            self.start_at: datetime = datetime.fromtimestamp(self.start_ut, timezone.utc)
            ("")
            self.end_ut: int = int(d['event_end'])
            ("")
            self.end_at: datetime = datetime.fromtimestamp(self.end_ut, timezone.utc)
            ("")
            self.is_live: bool = d['event_is_live']
            ("")

    class Flair:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.template_uuid: Optional[str] = d.get('link_flair_template_id')
            ("""
                The post flair template UUID.

                Value is null if no flair template is being used.
                """)
            self.text_mode: str = d['link_flair_type']
            ("""
                Either `text` or `richtext`.
                """)
            self.text: str = d['link_flair_text'] or ''
            ("""
                The flair text.
                """)
            self.css_class: str = d['link_flair_css_class'] or ''
            ("""
                Post flair CSS class.

                When a flair template is being used, the value of this field will be
                that of the CSS class designated by the template. If the flair template
                does not specify a CSS class then the value will be an empty string.
                """)
            # Post flairs can't be `transparent` but just in case.
            self.bg_color: str = '' if (x := d['link_flair_background_color']) == 'transparent' else x
            ("""A background color hex string. E.g., `#46d160`.

                If a flair template is not being used then the value is an empty string.

                A post flair template background color cannot be unset unlike user flair templates.
                """)
            self.fg_color_scheme: str = d['link_flair_text_color']
            ("""
                Either `dark` or `light`, or empty string.

                Value is empty string if a flair template is not being used.
                """)

    class Reports:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.ignoring: bool = d['ignore_reports']
            ("")
            self.num_reports: int = d['num_reports']
            ("")
            self.mod_reports: Sequence[ModReport] = [load_mod_report(m) for m in d['mod_reports']]
            ("")
            self.user_reports: Sequence[UserReport] = [load_user_report(m) for m in d['user_reports']]
            ("")

    class Edited:
        def __init__(self, outer: Submission) -> None:
            self.ut: int = outer.edited_ut
            ("")
            self.at: datetime = outer.edited_at
            ("")

    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.id36: str = d['id']
        ("""
            The ID of the submission as a base 36 number.
            """)
        self.id: int = int(self.id36, 36)
        ("")
        self.created_ut: int = int(d['created_utc'])
        ("""
            Unix timestamp of when the submission was made.
            """)
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("""
            When the submission was made.
            """)

        self.title: str = d['title']
        ("")
        self.score: int = d['score']
        ("""
            The number of upvotes (minus downvotes).
            """)
        self.score_hidden: bool = d['hide_score']
        ("""
            Whether the score should be hidden.
            """)
        self.comment_count: int = d['num_comments']
        ("""
            The number of comments.

            This should be treated as an estimation and may not match
            the number of actual visible comments. It could be higher or lower.
            """)

        self.permalink_path: str = d['permalink']
        ("""
            The URL of the submission, without the domain.
            """)
        self.permalink: str = AUTHORIZATION_BASE_URL + self.permalink_path
        ("""
            The URL of the submission.
            """)

        edited: Any = d['edited']
        self.is_edited: bool = bool(edited)
        ("""
            True if the submission was edited.
            """)
        self.edited_ut: int = int(edited) if edited else 0
        ("""
            Unix timestamp of when the submission was edited.

            Value is `0` if :attr:`is_edited` is false.
            """)
        self.edited_at: datetime = datetime.min
        ("""
            When the submission was edited.

            Value is `datetime.min` if :attr:`is_edited` is false.
            """)
        if self.is_edited:
            self.edited_at = datetime.fromtimestamp(self.edited_ut, timezone.utc)

        self.edited: Optional[Submission.Edited] = None
        ("""
            Value non-`None` if the comment was edited.
            """)
        if edited:
            self.edited = self.Edited(self)

        self.upvote_ratio: float = d['upvote_ratio']
        ("")
        self.removal_category: Optional[str] = d['removed_by_category']
        ("""
            `None` if not removed.

            If removed, possible values:

                `author`, `anti_evil_ops`, `community_ops`, `legal_operations`, `copyright_takedown`,
                `reddit`, `user`, `deleted`, `moderator`, `automod_filtered`.
            """)
        self.suggested_sort: str = d['suggested_sort'] or ''
        ("""
            Either: `confidence` (best), `new`, `old`, `top`, `qa`, `controversial`, or `live`.

            Value empty string if not set.
            """)
        self.stickied: bool = d['stickied']
        ("")
        self.archived: bool = d['archived']
        ("""
            Whether the post is archived.

            Archived posts cannot be commented on,
            but the author can still edit the OP.
            """)
        self.locked: bool = d['locked']
        ("")
        self.in_contest_mode: bool = d['contest_mode']
        ("""
            Whether the post is in contest mode.

            In contest mode, the comments are shown in a random order.
            """)
        self.nsfw: bool = d['over_18']
        ("")
        self.spoiler: bool = d['spoiler']
        ("")
        self.oc: bool = d['is_original_content']
        ("""
            Whether the post is marked as 'original content'.
            """)
        self.num_crossposts: int = d['num_crossposts']
        ("""
            Crosspost count.
            """)
        self.is_crosspostable: bool = d['is_crosspostable']
        ("""
            Whether the post can be crossposted.

            Will be `false` if the post was removed or deleted.
            """)
        self.is_robot_indexable: bool = d['is_robot_indexable']
        ("""
            Whether search engines should index this submission.

            Will be `false` if the post was removed or deleted.
            """)
        self.is_poster_profile_pinned: bool = d['pinned']
        ("""
            Whether the post is pinned to the poster's profile.

            This attribute can only be true if the submission object was obtained through a user listing.
            """)
        self.distinguished: str = d['distinguished'] or ''
        ("""
            Either `moderator` or `admin`, or empty string.

            Value empty string if not distinguished.
            """)

        self.event: Optional[Submission.Event] = None
        ("""
            Attributes related to event metadata of the submission.
            """)
        if 'event_start' in d:
            self.event = self.Event(d)

        self.me: Submission.Me = self.Me(d)
        ("""
            Attributes relating to the current user.

            If there is no user context, these values contain nonsense.
            """)

        self.subreddit: Submission.Subreddit = self.Subreddit(d)
        ("""
            Information related to the submission's subreddit.
            """)

        author: str = d['author']
        self.author_display_name: str = author
        ("""
            The author's username.

            Possibly `[removed]` if the submission was removed or
            `[deleted]` if the submission was deleted by the author.
            """)
        self.author: Optional[Submission.Author] = None
        ("""
            Information about the author.

            Value is `None` if the submission was removed or deleted.
            """)
        if not author.startswith('['):
            self.author = self.Author(d)

        self.mod: Optional[Submission.Moderation] = None
        ("""
            Attributes relating to moderation.

            Value is `None` if the current user is not a moderator of the subreddit.
            """)
        if 'spam' in d:
            self.mod = self.Moderation(d)

        self.flair: Submission.Flair = self.Flair(d)
        ("""
            Attributes related to the author's flair.
            """)

        self.reports: Optional[Submission.Reports] = None
        ("")
        if d['num_reports'] is not None:
            self.reports = self.Reports(d)


class LinkPost(Submission):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.link: str = d['url_overridden_by_dest']
        ("""
            The URL linked to by the link post.
            """)

class TextPost(Submission):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.body: str = d['selftext']
        ("""
            The body text of the submission. In markdown format.
            """)
        self.body_html: str = d['selftext_html']
        ("""
            The HTML of the submission.
            """)

class GalleryPost(Submission):
    @dataclass(repr=False, eq=False)
    class GalleryItem:
        id: int
        media_id: str
        caption: str
        outbound_link: str

    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.gallery_link: str = d['url_overridden_by_dest']
        ("")

        gallery_data_items: Sequence[Any] = ()
        if gallery_data := d.get('gallery_data'):
            gallery_data_items = gallery_data['items']
        self.gallery: Sequence[GalleryPost.GalleryItem] = [
            self.GalleryItem(
                id=m['id'],
                media_id=m['media_id'],
                caption=m.get('caption', ''),
                outbound_link=m.get('outbound_url', ''),
            )
            for m in gallery_data_items
        ]
        ("")

class PollPost(Submission):
    pass

class CrosspostSubmission(Submission):
    @property
    def original(self) -> Optional[Submission]:
        """Original submission of this crosspost.

        Value `None` if the original submission is from a subreddit that is
        now banned and the submission is no longer accessible.
        """
        return self.__original

    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.original_id36: str = d['crosspost_parent'][3:]
        ("")
        self.original_id: int = int(self.original_id36, 36)
        ("")

        self.__original: Optional[Submission] = None
        # https://github.com/python/mypy/issues/4177
        if self.__class__.original == __class__.original:  # type: ignore[name-defined]
            from ..model_loaders.submission import load_submission  # Avoid cyclic import
            self.__original = self._load_original(d, load_submission)

    _TSubmission = TypeVar('_TSubmission', bound=Submission)

    @final
    def _load_original(self,
        d: Mapping[str, Any],
        load: Callable[[Mapping[str, Any]], _TSubmission],
    ) -> Optional[_TSubmission]:
        if crosspost_parent_list := d['crosspost_parent_list']:
            return load(crosspost_parent_list[0])
        return None




from __future__ import annotations
from typing import TypeVar, AsyncIterator, Callable, Optional, Awaitable, Iterable

T = TypeVar('T')

class StubbornCallerAsyncIterator(AsyncIterator[T]):
    """Call each callable in the given iterator and return its result.

    If a call raises an exception it will propagate normally. Doing
    `next(self)` will re-attempt the call until it returns a result.

    Has a `self.current` attribute to get the current callable.
    """

    def __init__(self, iterable: Iterable[Callable[[], Awaitable[T]]]) -> None:
        self.__itr = iter(iterable)
        self.current: Optional[Callable[[], Awaitable[T]]] = None
        ("")

    def __aiter__(self) -> AsyncIterator[T]:
        return self

    async def __anext__(self) -> T:
        if self.current is None:
            try:
                self.current = next(self.__itr)
            except StopIteration:
                raise StopAsyncIteration
        ret = await self.current()
        self.current = None
        return ret




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Callable, MutableMapping
if TYPE_CHECKING:
    from ..auth.token_obtainment_client_ASYNC import TokenObtainmentClient
    from ..auth.token import Token
    from ...http.handler_ASYNC import Handler
    from ...http.requisition import Requisition
    from ...http.send_params import SendParams
    from ...http.exchange import Exchange

from ...util.imports import lazy_import
if TYPE_CHECKING:
    import asyncio
else:
    lazy_import % 'asyncio'

import time

from ...http.delegating_handler_ASYNC import DelegatingHandler


class Authorizer:
    def __init__(self,
        token_client: Optional[TokenObtainmentClient] = None,
        token: Optional[Token] = None,
    ) -> None:
        self._token_client: Optional[TokenObtainmentClient] = token_client
        ("")
        self._token: Optional[Token] = token
        ("")
        self.renewal_time: Optional[float] = None
        ("")
        self.renewal_skew: float = 30
        ("")
        self.expires_in_fallback: Optional[int] = None
        ("")
        self.time_func: Callable[[], float] = time.monotonic
        ("")
        self.authorization_header_name: str = 'Authorization'
        ("")

    def has_token_client(self) -> bool:
        return self._token_client is not None

    def fetch_token_client(self) -> TokenObtainmentClient:
        v = self._token_client
        if v is None:
            raise RuntimeError('token client not set')
        return v

    def set_token_client(self, value: Optional[TokenObtainmentClient]) -> None:
        self._token_client = value

    def has_token(self) -> bool:
        return self._token is not None

    def fetch_token(self) -> Token:
        v = self._token
        if v is None:
            raise RuntimeError('token not set')
        return v

    def set_token(self, value: Optional[Token]) -> None:
        self._token = value

    async def renew_token(self) -> None:
        tc = self.fetch_token_client()
        tk = await tc.fetch_token()

        self.set_token(tk)

        expires_in = tk.expires_in
        if expires_in is None:
            expires_in = self.expires_in_fallback
        if expires_in is None:
            self.renewal_time = None
        else:
            self.renewal_time = self.time() + expires_in - self.renewal_skew

    def time(self) -> float:
        return self.time_func()

    def should_renew_token(self) -> bool:
        if not self.has_token():
            return True
        if self.renewal_time is None:
            return False
        return self.time() >= self.renewal_time

    def can_renew_token(self) -> bool:
        return self.has_token_client()

    async def attain_token(self) -> Token:
        if self.should_renew_token():
            await self.renew_token()
        return self.fetch_token()

    def get_header_entry(self) -> tuple[str, str]:
        return (self.authorization_header_name, 'Bearer {0.access_token}'.format(self.fetch_token()))

    def prepare_requisition(self, requisition: Requisition) -> None:
        self.prepare_headers(requisition.headers)

    def prepare_headers(self, headers: MutableMapping[str, str]) -> None:
        k, v = self.get_header_entry()
        headers[k] = v


class Authorized(DelegatingHandler):
    def __init__(self, handler: Handler, authorizer: Authorizer) -> None:
        super().__init__(handler)
        self.authorizer: Authorizer = authorizer
        ("")
        self._lock = asyncio.Lock()

    async def _send(self, p: SendParams) -> Exchange:
        authorizer = self.authorizer
        async with self._lock:
            if authorizer.should_renew_token():
                await authorizer.renew_token()

        used_token = authorizer.fetch_token()
        authorizer.prepare_requisition(p.requisition)

        xchg = await super()._send(p)

        maybe_invalid_token = xchg.response.status == 401
        if maybe_invalid_token and authorizer.can_renew_token():
            async with self._lock:
                if used_token.access_token == authorizer.fetch_token().access_token:
                    await authorizer.renew_token()

            authorizer.prepare_requisition(p.requisition)

            xchg = await super()._send(p)

        return xchg




from __future__ import annotations
from typing import Sequence

from .listing.listing_async_paginator import ListingAsyncPaginator
from ...models.message_ASYNC import MailboxMessage, ComposedMessage, CommentMessage
from ...model_loaders.message_ASYNC import (
    load_mailbox_message,
    load_composed_message,
    load_comment_message,
    load_composed_message_thread,
)

class MessageListingAsyncPaginator(ListingAsyncPaginator[MailboxMessage]):
    async def fetch(self) -> Sequence[MailboxMessage]:
        data = await self._fetch_data()
        return [load_mailbox_message(d['data'], self.client) for d in data['children']]

class ComposedMessageListingAsyncPaginator(ListingAsyncPaginator[ComposedMessage]):
    async def fetch(self) -> Sequence[ComposedMessage]:
        data = await self._fetch_data()
        return [load_composed_message(d['data'], self.client) for d in data['children']]

class CommentMessageListingAsyncPaginator(ListingAsyncPaginator[CommentMessage]):
    async def fetch(self) -> Sequence[CommentMessage]:
        data = await self._fetch_data()
        return [load_comment_message(d['data'], self.client) for d in data['children']]

class ComposedMessageThreadListingAsyncPaginator(ListingAsyncPaginator[Sequence[ComposedMessage]]):
    async def fetch(self) -> Sequence[Sequence[ComposedMessage]]:
        data = await self._fetch_data()
        return [load_composed_message_thread(d['data'], self.client) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.message_ASYNC import ComposedMessage

from functools import cached_property

from ...util.base_conversion import to_base36
from ...model_loaders.message_ASYNC import load_composed_message, load_composed_message_thread
from .pulls_ASYNC import Pulls

class MessageProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.pulls: Pulls = Pulls(client)
        ("""
            Pull messages.
            """)

    async def get_thread(self, idn: int) -> Sequence[ComposedMessage]:
        """Get a message thread.

        Specifying the ID of any message in the same thread gives you the same list.

        .. .PARAMETERS

        :param `int` idn:
            Message ID.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.message_ASYNC.ComposedMessage`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The target message specified does not exist or
                you do not have permission to access it.
        """
        id36 = to_base36(idn)
        root = await self._client.request('GET', f'/message/messages/{id36}')
        obj_data = root['data']['children'][0]['data']
        return load_composed_message_thread(obj_data, self._client)

    async def send(self, to: str, subject: str, body: str) -> None:
        """Send a direct message to a user.

        .. .PARAMETERS

        :param `str` to:
            Name of user in which to send the message to.
        :param `str` subject:
            The subject of the message.
        :param `str` body:
            The body text of the message.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The user specified by `to` was empty.
            + `NO_SUBJECT`:
                The subject specified by `subject` was empty.
            + `NO_TEXT`:
                The body text specified by `body` was empty.
            + `USER_DOESNT_EXIST`:
                The user specified by `to` does not exist.
            + `NOT_WHITELISTED_BY_USER_MESSAGE`:
                The target user has direct messages from strangers disabled.

                On old Reddit, this setting can be configured at `<https://old.reddit.com/prefs/blocked>`_
                by setting "Show private messages from" to "Only trusted users".
        """
        req_data = {
            'to': to,
            'subject': subject,
            'text': body,
        }
        await self._client.request('POST', '/api/compose', data=req_data)

    async def send_from_sr(self, sr: str, to: str, subject: str, body: str) -> None:
        """Send a direct message to a user on behalf of a subreddit.

        .. .PARAMETERS

        :param `str` sr:
            The name of a subreddit you moderate.
        :param `str` to:
            Name of user in which to send the message to.
        :param `str` subject:
            The subject of the message.
        :param `str` body:
            The body text of the message.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The user specified by `to` was empty.
            + `NO_SUBJECT`:
                The subject specified by `subject` was empty.
            + `NO_TEXT`:
                The body text specified by `body` was empty.
            + `SUBREDDIT_NOEXIST`:
                The subreddit specified by `sr` does not exist.
            + `NO_SR_TO_SR_MESSAGE`:
                Both `sr` and `to` refer to subreddits.
                Subreddit to subreddit messages is not allowed.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The current user is not a moderator of the specified subreddit.
        """
        req_data = {
            'from_sr': sr,
            'to': to,
            'subject': subject,
            'text': body,
        }
        await self._client.request('POST', '/api/compose', data=req_data)

    async def reply(self, idn: int, body: str) -> ComposedMessage:
        """Reply to a message

        .. .PARAMETERS

        :param `int` idn:
        :param `str` body:

        .. .RETURNS

        :returns: The newly created message.
        :rtype: :class:`~.models.message_ASYNC.ComposedMessage`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The body text specified by `body` was empty.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The specified message ID does not exist.
        """
        data = {
            'thing_id': 't4_' + to_base36(idn),
            'text': body,
            'return_rtjson': '1',
        }
        result = await self._client.request('POST', '/api/comment', data=data)
        root = result['json']['data']['things'][0]['data']
        return load_composed_message(root, self._client)

    async def delete(self, idn: int) -> None:
        """Delete a message.

        If the message specified by the ID doesn't exist, or is invalid,
        the action is treated as a success.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        await self._client.request('POST', '/api/del_msg', data={'id': 't4_' + to_base36(idn)})

    async def mark_read(self, idn: int) -> None:
        """Mark a message as read.

        Marking an already marked as read item is treated as a success.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
                The message specified by the ID doesn't exist or is invalid.
        """
        await self._client.request('POST', '/api/read_message', data={'id': 't4_' + to_base36(idn)})

    async def mark_unread(self, idn: int) -> None:
        """Mark a message as unread.

        Behaves similarly to :meth:`.mark_read`.
        """
        await self._client.request('POST', '/api/unread_message', data={'id': 't4_' + to_base36(idn)})

    async def mark_all_read(self) -> None:
        """Mark all messages as read.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        await self._client.request('POST', '/api/read_all_messages')

    async def mark_comment_read(self, idn: int) -> None:
        """Mark a comment message as read.

        Behaves similarly to :meth:`.mark_read`.
        """
        await self._client.request('POST', '/api/read_message', data={'id': 't1_' + to_base36(idn)})

    async def mark_comment_unread(self, idn: int) -> None:
        """Mark a comment message as unread.

        Behaves similarly to :meth:`.mark_read`.
        """
        await self._client.request('POST', '/api/read_message', data={'id': 't1_' + to_base36(idn)})

    async def collapse(self, idn: int) -> None:
        """Collapse a message.

        If the specified message does not exist or is valid,
        the action is treated as a success.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        await self._client.request('POST', '/api/collapse_message', data={'id': 't4_' + to_base36(idn)})

    async def uncollapse(self, idn: int) -> None:
        """Uncollapse a message.

        Behaves similarly to :meth:`.mark_read`.
        """
        await self._client.request('POST', '/api/uncollapse_message', data={'id': 't4_' + to_base36(idn)})

    class BlockAuthor:
        def __init__(self, outer: MessageProcedures) -> None:
            self._client = outer._client

        async def __call__(self, idn: int) -> None:
            await self.of_message(idn)

        async def of_message(self, idn: int) -> None:
            await self._client.request('POST', '/api/block', data={'id': 't4_' + to_base36(idn)})

        async def of_comment(self, idn: int) -> None:
            await self._client.request('POST', '/api/block', data={'id': 't1_' + to_base36(idn)})

        async def of_submission(self, idn: int) -> None:
            await self._client.request('POST', '/api/block', data={'id': 't3_' + to_base36(idn)})

    block_author: cached_property[BlockAuthor] = cached_property(BlockAuthor)
    ("""
        Block the author of a message, comment, or submission.

        - Use :meth:`~.BlockAuthor.__call__` to block the author of a message.
        - Use :meth:`~.BlockAuthor.of_comment` to block the author of a comment.
        - Use :meth:`~.BlockAuthor.of_submission` to block the author of a submission.

        To block a user directly by ID or name, see
        :meth:`~.AccountProcedures.block_user_by_id` and :meth:`~.AccountProcedures.block_user_by_name`
        instead.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """)




from __future__ import annotations
from typing import Any, Optional, Sequence, Mapping

from datetime import datetime
from dataclasses import dataclass
from enum import IntEnum

from .datamemento import DatamementoPropertiesMixin


class ConversationProgress(IntEnum):
    NEW = 0
    IN_PROGRESS = 1
    ARCHIVED = 2

class ModActionType(IntEnum):
    HIGHLIGHT = 0
    UNHIGHLIGHT = 1
    ARCHIVE = 2
    UNARCHIVE = 3
    MUTE_USER = 5
    UNMUTE_USER = 6
    BAN_USER = 7
    UNBAN_USER = 8
    APPROVE_USER = 9
    DISAPPROVE_USER = 10


@dataclass(repr=False, eq=False)
class ModmailSubreddit(DatamementoPropertiesMixin):
    d: Mapping[str, Any]
    id: int
    name: str
    subscriber_count: int


class ConversationInfo(DatamementoPropertiesMixin):
    @dataclass(repr=False, eq=False)
    class LegacyMessage:
        id: int
        id36: str
        link: str

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id36: str = d['id']
        ("")
        self.id: int = int(self.id36, 36)
        ("")
        self.subject: str = d['subject']
        ("")
        self.progress: int = d['state']
        ("""
            Enum: :class:`.ConversationProgress`
            """)
        self.message_count: int = d['numMessages']
        ("""
            Number of messages in the conversation.
            """)
        self.auto: bool = d['isAuto']
        ("""
            True if the conversation was created due to an automated message;
            false if the message was created by a user.
            """)
        self.internal: bool = d['isInternal']
        ("""
            Whether this conversation is a moderator discussion.
            """)
        self.repliable: bool = d['isRepliable']
        ("""
            True if the conversation thread accepts replies.
            """)
        self.highlighted: bool = d['isHighlighted']
        ("""
            True if the conversation is highlighted.
            """)

        self.legacy_message: Optional[ConversationInfo.LegacyMessage] = None
        ("""
            Object that contains an ID that refers to the first message in the
            legacy modmail thread for this conversation thread.

            The underlying data can be null in rare cases, such as in the
            "r/{subreddit} is now enrolled in the New Modmail"
            modmail message from u/reddit.
            """)
        if (x := d['legacyFirstMessageId']) is not None:
            self.legacy_message = self.LegacyMessage(
                id36=x,
                id=int(x, 36),
                link="https://www.reddit.com/message/messages/" + x,
            )

        self.last_updated_by_user_at: Optional[datetime] = (x := d['lastUserUpdate']) and datetime.fromisoformat(x)
        ("""
            If :attr:`internal` is true this should always be null.
            """)
        self.last_updated_by_mod_at: Optional[datetime] = (x := d['lastModUpdate']) and datetime.fromisoformat(x)
        ("")
        self.last_updated_at: datetime = datetime.fromisoformat(d['lastUpdated'])
        ("""
            Same as either :attr:`last_updated_by_user_at` or :attr:`last_updated_by_mod_at`, whichever is newer.
            """)
        self.last_unread: Optional[datetime] = (x := d['lastUnread']) and datetime.fromisoformat(x)
        ("""
            Datetime of when the conversation was last marked unread.
            """)
        owner = d['owner']
        self.subreddit_name: str = owner['displayName']
        ("""
            Name of the subreddit associated with this conversation.
            """)
        self.subreddit_id: int = int(owner['id'][3:], 36)
        ("""
            ID of the subreddit associated with this conversation.
            """)

class Message(DatamementoPropertiesMixin):
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id36: str = d['id']
        ("")
        self.id: int = int(self.id36, 36)
        ("")
        author = d['author']
        self.author_name: str = author['name']
        ("""
            User name of the author of the message.

            Unknown what happens if the user is deleted.
            Is the value `[deleted]`, an empty string, or does the field in the
            underlying object not exist?
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        self.author_id: int = author['id']
        ("""
            User ID of the author of the message.

            Unknown what happens if the user is deleted.
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        self.body: str = d['bodyMarkdown']
        ("""
            Text content of the message in markdown format.
            """)
        self.body_html: str = d['body']
        ("""
            Content of the message in HTML.
            """)
        self.datetime: datetime = datetime.fromisoformat(d['date'])
        ("""
            Datetime object of when the message was created.
            """)
        self.internal: bool = d['isInternal']
        ("""
            Always true if this message is in a moderator discussion.
            Otherwise, true if this message is a private moderator note.
            """)

class ModAction(DatamementoPropertiesMixin):
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id: int = int(d['id'], 36)
        ("")
        self.action_type: int = d['actionTypeId']
        ("""
            Enum: :class:`.ModActionType`
            """)
        self.agent_name: str = d['author']['name']
        ("""
            Name of the mod who performed the action.

            Unknown what happens if the user is deleted.
            Is the value `[deleted]`, an empty string, or does the field in the
            underlying object not exist?
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        self.agent_id: int = d['author']['id']
        ("""
            User ID of the mod who performed the action.

            Unknown what happens if the user is deleted.
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        self.datetime: datetime = datetime.fromisoformat(d['date'])
        ("""
            Datetime object of when the action was performed.
            """)

class UserDossier(DatamementoPropertiesMixin):
    @dataclass(repr=False, eq=False)
    class RecentPost:
        id: int
        ("""
            The submission ID.
            """)
        permalink: str
        ("""
            A URL to the submission.
            """)
        title: str
        ("""
            Title of the submission.
            """)
        created_at: datetime
        ("""
            When the submission was created.
            """)

    @dataclass(repr=False, eq=False)
    class RecentComment:
        id: int
        ("""
            The comment ID.
            """)
        permalink: str
        ("""
            A URL to the comment.
            """)
        submission_title: str
        ("""
            The title of the submission in which the comment resides.
            """)
        body: str
        ("""
            The comment the user wrote.
            """)
        created_at: datetime
        ("""
            When the comment was created.
            """)

    @dataclass(repr=False, eq=False)
    class RecentConvo:
        id: int
        ("""
            The conversation ID.
            """)
        subject: str
        ("""
            The subject line of the conversation.
            """)
        permalink: str
        ("""
            A URL to the conversation.

            E.g., https://mod.reddit.com/mail/perma/tiebu.
            """)

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id: int = int(d['id'][3:], 36)
        ("")
        self.username: str = d['name']
        ("""
            The name of the target user.
            """)
        self.created_at: datetime = datetime.fromisoformat(d['created'])
        ("""
            When the user account was created.
            """)
        self.suspended: bool = d['isSuspended']
        ("""
            True if account is suspended.
            """)
        self.shadow_banned: bool = d['isShadowBanned']
        ("""
            True if account is shadow banned.
            """)
        self.approved_user: bool = d['approveStatus']['isApproved']
        ("""
            True if the user is an approved contributor on the subreddit.
            """)

        mute_status = d['muteStatus']
        self.muted: bool = mute_status['isMuted']
        ("""
            True if the user is currently muted on the subreddit.
            """)
        self.mute_reason: str = mute_status['reason']
        ("""
            Mute reason.

            Empty string if not currently muted.
            """)
        self.mute_count: int = mute_status['muteCount']
        ("""
            Number of times the user has been muted in the subreddit.
            """)
        self.mute_end_datetime: datetime = datetime.min if (x := mute_status['endDate']) is None else datetime.fromisoformat(x)
        ("""
            Datetime object of when the mute ends.

            Value is `datetime.min` if :attr:`muted` is false.
            """)

        ban_status = d['banStatus']
        self.banned: bool = ban_status['isBanned']
        ("""
            True if the user is currently banned on the subreddit.
            """)
        self.ban_reason: str = ban_status['reason']
        ("""
            Ban reason.

            Empty string if not currently banned.
            """)
        self.ban_permanent: bool = ban_status['isPermanent']
        ("""
            True if the ban is permanent.

            Value false if user is not banned.
            """)
        self.ban_end_datetime: datetime = datetime.min if (x := ban_status['endDate']) is None else datetime.fromisoformat(x)
        ("""
            Datetime object of when the ban ends.

            Value is `datetime.min` if :attr:`banned` is false.
            """)

        self.recent_posts: Sequence[UserDossier.RecentPost] = [
            self.RecentPost(
                id=int(k.partition('_')[-1], 36),
                permalink=v['permalink'],
                title=v['title'],
                created_at=datetime.fromisoformat(v['date']),
            )
            for k, v in d['recentPosts'].items()
        ]
        ("""
            A bit of information about the user's recent submissions to the subreddit.
            """)
        self.recent_comments: Sequence[UserDossier.RecentComment] = [
            self.RecentComment(
                id=int(k.partition('_')[-1], 36),
                permalink=v['permalink'],
                submission_title=v['title'],
                body=v['comment'],
                created_at=datetime.fromisoformat(v['date']),
            )
            for k, v in d['recentComments'].items()
        ]
        ("""
            A bit information about the user's recent comments in the subreddit.
            """)
        self.recent_convos: Sequence[UserDossier.RecentConvo] = [
            self.RecentConvo(
                id=int(v['id'], 36),
                subject=v['subject'],
                permalink=v['permalink'],
            )
            for _k, v in d['recentConvos'].items()
        ]
        ("""
            Other modmail conversations this user is involved in.
            """)


@dataclass(repr=False, eq=False, frozen=True)
class ConversationAggregate:
    info: ConversationInfo
    ("""
        Information about the conversation.
        """)
    messages: Sequence[Message]
    ("""
        Conversation messages.
        """)
    mod_actions: Sequence[ModAction]
    ("""
        Conversation mod actions.
        """)
    history: Sequence[object]
    ("""
        Conversation entries.

        Objects are either :attr:`.Message` or :attr:`.ModAction` instances.
        """)

@dataclass(repr=False, eq=False, frozen=True)
class UserDossierConversationAggregate(ConversationAggregate):
    user_dossier: UserDossier
    ("""
        Information about the target user.
        """)

@dataclass(repr=False, eq=False, frozen=True)
class OptionalUserDossierConversationAggregate(ConversationAggregate):
    user_dossier: Optional[UserDossier]
    ("""
        Information about the target user.
        """)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping, Sequence
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from ..models.message_ASYNC import MailboxMessage, ComposedMessage, CommentMessage
from .message import (
    load_composed_message as load_base_composed_message,
    load_comment_message as load_base_comment_message,
)

def load_composed_message(d: Mapping[str, Any], client: Client) -> ComposedMessage:
    up = load_base_composed_message(d)
    return ComposedMessage(
        d=up.d,
        subject=up.subject,
        author_name=up.author_name,
        unread=up.unread,
        id=up.id,
        unixtime=up.unixtime,
        datetime=up.datetime,
        body=up.body,
        body_html=up.body_html,
        distinguished=up.distinguished,
        src_user_name=up.src_user_name,
        src_subr_name=up.src_subr_name,
        dst_user_name=up.dst_user_name,
        dst_subr_name=up.dst_subr_name,
        src_user_id=up.src_user_id,
        client=client,
    )

def load_comment_message(d: Mapping[str, Any], client: Client) -> CommentMessage:
    up = load_base_comment_message(d)
    return CommentMessage(
        d=up.d,
        subject=up.subject,
        author_name=up.author_name,
        unread=up.unread,
        cause=up.cause,
        submission=up.submission,
        comment=up.comment,
        client=client,
    )

def load_mailbox_message(d: Mapping[str, Any], client: Client) -> MailboxMessage:
    if d['was_comment']:
        return load_comment_message(d, client)
    return load_composed_message(d, client)


def load_composed_message_thread(d: Mapping[str, Any], client: Client) -> Sequence[ComposedMessage]:
    first = load_composed_message(d, client)
    children = []
    if replies := d['replies']:
        children = [load_composed_message(d['data'], client) for d in replies['data']['children']]
    return [first] + children




from __future__ import annotations
from typing import Sequence

from .listing.listing_async_paginator import ListingAsyncPaginator
from ...models.live_thread_ASYNC import LiveUpdate
from ...model_loaders.live_thread_ASYNC import load_live_update

class LiveUpdateListingAsyncPaginator(ListingAsyncPaginator[LiveUpdate]):
    async def fetch(self) -> Sequence[LiveUpdate]:
        data = await self._fetch_data()
        return [load_live_update(d['data'], self.client) for d in data['children']]




from __future__ import annotations

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class CommunityListWidgetItem:
    name: str
    subscribers: int
    icon_img: str
    community_icon: str
    primary_color: str
    nsfw: bool




from .pull_async1 import (  # noqa: F401
    ModQueueListingAsyncPaginator as ModQueueListingAsyncPaginator,
    ModQueueSubmissionListingAsyncPaginator as ModQueueSubmissionListingAsyncPaginator,
    ModQueueCommentListingAsyncPaginator as ModQueueCommentListingAsyncPaginator,
    ReportsListingAsyncPaginator as ReportsListingAsyncPaginator,
    ReportsSubmissionListingAsyncPaginator as ReportsSubmissionListingAsyncPaginator,
    ReportsCommentListingAsyncPaginator as ReportsCommentListingAsyncPaginator,
    SpamListingAsyncPaginator as SpamListingAsyncPaginator,
    SpamSubmissionListingAsyncPaginator as SpamSubmissionListingAsyncPaginator,
    SpamCommentListingAsyncPaginator as SpamCommentListingAsyncPaginator,
    EditedListingAsyncPaginator as EditedListingAsyncPaginator,
    EditedSubmissionListingAsyncPaginator as EditedSubmissionListingAsyncPaginator,
    EditedCommentListingAsyncPaginator as EditedCommentListingAsyncPaginator,
    UnmoderatedSubmissionListingAsyncPaginator as UnmoderatedSubmissionListingAsyncPaginator,
)
from .pull_users_async1 import (  # noqa: F401
    ModeratorsAsyncPaginator as ModeratorsAsyncPaginator,
    ApprovedUsersAsyncPaginator as ApprovedUsersAsyncPaginator,
    BannedUsersAsyncPaginator as BannedUsersAsyncPaginator,
    MutedUsersAsyncPaginator as MutedUsersAsyncPaginator,
)
from .legacy_pull_users_async1 import (  # noqa: F401
    UserRelationshipListingAsyncPaginator as UserRelationshipListingAsyncPaginator,
    BannedSubredditUserRelationshipListingAsyncPaginator as BannedSubredditUserRelationshipListingAsyncPaginator,
)
from .pull_actions_async1 import ModerationActionLogAsyncPaginator as ModerationActionLogAsyncPaginator  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
if TYPE_CHECKING:
    from .moderator_list import ModeratorInfo

from dataclasses import dataclass

from .widget_ASYNC import Widget

@dataclass(repr=False, eq=False)
class ModeratorListWidget(Widget):
    mod_num: int
    mods: Sequence[ModeratorInfo]




from __future__ import annotations
from typing import Mapping, Any

from datetime import datetime, timezone

from .datamemento import DatamementoBase

class MyAccount(DatamementoBase):
    class Subreddit:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.name: str = d['display_name']
            ("")
            self.openness: str = d['subreddit_type']
            ("""
                Either: `public`, `private`, `restricted`, `archived`,
                `employees_only`, `gold_only`, `gold_restricted`, or `user`.
                """)
            self.subscriber_count: int = d['subscribers']
            ("")
            self.title: str = d['title']
            ("")
            self.public_description: str = d['public_description']
            ("")
            self.nsfw: bool = d['over_18']
            ("")

    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.id36: str = d['id']
        ("""
            User ID of the current user account as a base 36 number.
            """)
        self.id: int = int(self.id36, 36)
        ("""
            User ID of the current user account.
            """)
        self.name: str = d['name']
        ("""
            Username of the current user account.
            """)
        self.created_ut: int = int(d['created_utc'])
        ("""
            Unix timestamp of when the current user account was created.
            """)
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("""
            Datetime object of when the current user account was created.
            """)
        self.post_karma: int = d['link_karma']
        ("")
        self.comment_karma: int = d['comment_karma']
        ("")
        self.awardee_karma: int = d['awardee_karma']
        ("")
        self.awarder_karma: int = d['awarder_karma']
        ("")
        self.total_karma: int = d['total_karma']
        ("")
        self.has_mail: bool = d['has_mail']
        ("""
            True if the current user has a new inbox message.
            """)
        self.has_mod_mail: bool = d['has_mod_mail']
        ("""
            True if the current user has a new modmail message.
            """)
        self.inbox_count: int = d['inbox_count']
        ("""
            Number of new inbox messages.
            """)
        self.coins: int = d['coins']
        ("""
            How rich the current user is.
            """)
        self.friend_count: int = d['num_friends']
        ("""
            Number of friends in the current user's friends list.

            This is an old reddit feature. See `<https://www.reddit.com/prefs/friends/>`_.
            """)
        self.nsfw: bool = d['over_18']
        ("""
            Whether the current user account has opted in to NSFW content.

            On old reddit, this is the preference option that says
            "I am over eighteen years old and willing to view adult content".
            """)

        self.subreddit: MyAccount.Subreddit = self.Subreddit(d['subreddit'])
        ("""
            Information about the current user account's user subreddit.
            """)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .send_params import SendParams
    from .exchange import Exchange

from .handler_ASYNC import Handler


class DelegatingHandler(Handler):
    def __init__(self, handler: Handler) -> None:
        self._handler: Handler = handler

    async def _send(self, p: SendParams) -> Exchange:
        return await self._handler._send(p)

    async def _close(self) -> None:
        await self._handler._close()




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.modmail import ModmailSubreddit

from ...model_loaders.modmail import load_modmail_subreddit
from .conversation_ASYNC import ConversationProcedures
from .pull_ASYNC import Pull

class ModmailProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.conversation: ConversationProcedures = ConversationProcedures(client)
        self.pull: Pull = Pull(client)

    async def get_unread_counts(self) -> Mapping[str, int]:
        """Get unread conversations counts by mailbox.

        Returns a dictionary like the following::

           {"archived": 0,
            "appeals": 0,
            "highlighted": 0,
            "notifications": 2,
            "join_requests": 0,
            "filtered": 0,
            "new": 1,
            "inprogress": 0,
            "mod": 0}

        .. .RETURNS

        :rtype: `Mapping`\\[`str`, `int`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        return await self._client.request('GET', '/api/mod/conversations/unread/count')

    async def subreddits(self) -> Sequence[ModmailSubreddit]:
        """Return subreddits the current user is moderating that have modmail enabled.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.modmail.ModmailSubreddit`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `500`:
                There is no user context.
        """
        root = await self._client.request('GET', '/api/mod/conversations/subreddits')
        return [load_modmail_subreddit(d) for d in root['subreddits'].values()]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from .custom_css import CustomCSSWidgetImageInfo

from dataclasses import dataclass

from .widget_SYNC import Widget

@dataclass(repr=False, eq=False)
class CustomCSSWidget(Widget):
    text: str
    css: str
    height: Optional[int]
    image_data: Sequence[CustomCSSWidgetImageInfo]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union
if TYPE_CHECKING:
    from types import ModuleType
    from importlib.machinery import ModuleSpec
    from os import PathLike

import sys
import importlib.util


def load_spec(name: str, package: Optional[str] = None) -> ModuleSpec:
    spec = importlib.util.find_spec(name, package)
    if spec is None:
        raise RuntimeError(f'module spec not found: {name} ({package})')
    return spec

def load_spec_from_file_location(name: str,
        location: Union[str, bytes, PathLike[str], PathLike[bytes]]) -> ModuleSpec:
    spec = importlib.util.spec_from_file_location(name, location)
    if spec is None:
        raise RuntimeError(f'module spec not found: {name} ({str(location)})')
    return spec

def import_module_from_spec(spec: ModuleSpec) -> ModuleType:
    if spec.loader is None:
        raise RuntimeError('spec has no loader')
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


class LazyImport:
    """An object for lazily loading modules."""

    def __call__(self, name: str, package: Optional[str] = None) -> ModuleType:
        try:
            return sys.modules[name]
        except KeyError:
            spec = importlib.util.find_spec(name, package)
            if spec is None:
                raise RuntimeError(f'module named {name!r} not found')
            if spec.loader is None:
                raise RuntimeError('spec has no loader')
            loader = importlib.util.LazyLoader(spec.loader)
            spec.loader = loader
            return import_module_from_spec(spec)

    def __mod__(self, other: str) -> None:
        if '.' in other:
            raise ValueError('dotted module name not supported')

        module = self(other)
        caller_frame = sys._getframe(1)  # type: ignore
        caller_frame.f_globals[other] = module

lazy_import: LazyImport = LazyImport()




from __future__ import annotations
from typing import TypeVar, Iterable, Iterator, Sequence
from itertools import islice

T = TypeVar('T')

def chunked(src: Iterable[T], size: int) -> Iterator[Sequence[T]]:
    itr = iter(src)
    while chunk := tuple(islice(itr, size)):
        yield chunk

class ChunkingIterator(Iterator[Sequence[T]]):
    """Chunking iterator.

    The chunk size can be controlled during iteration using `.size`.
    """

    def __init__(self, src: Iterable[T], size: int) -> None:
        self._itr = iter(src)
        self.size: int = size
        ("")

    def __iter__(self) -> Iterator[Sequence[T]]:
        return self

    def __next__(self) -> Sequence[T]:
        if chunk := tuple(islice(self._itr, self.size)):
            return chunk
        raise StopIteration




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from .subreddit import (
    Subreddit as BaseSubreddit,
    InaccessibleSubreddit as BaseInaccessibleSubreddit,
)

class Subreddit(BaseSubreddit):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

Community = Subreddit

class InaccessibleSubreddit(BaseInaccessibleSubreddit):
    def __init__(self, d: Mapping[str, Any], client: Client) -> None:
        super().__init__(d)
        self.client: Client = client
        ("")

InaccessibleCommunity = InaccessibleSubreddit




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, MutableSequence
if TYPE_CHECKING:
    from ..http.handler_SYNC import Handler
    from ..http.send_params import SendParams
    from ..http.exchange import Exchange
    from ..http.requisition import Requisition
    from ..http.request import Request
    from ..http.response import Response

from collections import deque

from ..http.delegating_handler_SYNC import DelegatingHandler

class Recorded(DelegatingHandler):
    def __init__(self, handler: Handler) -> None:
        super().__init__(handler)
        self.last_requisition: Optional[Requisition] = None
        ("")
        self.last_exchange: Optional[Exchange] = None
        ("")
        self.last_transmit: Optional[tuple[Requisition, Optional[Exchange]]] = None
        ("")
        self.last_transfer: Optional[tuple[Requisition, Exchange]] = None
        ("")
        self.last_requisition_queue: MutableSequence[Requisition] = deque(maxlen=16)
        ("")
        self.last_exchange_queue: MutableSequence[Exchange] = deque(maxlen=16)
        ("")
        self.last_transmit_queue: MutableSequence[tuple[Requisition, Optional[Exchange]]] = deque(maxlen=16)
        ("")
        self.last_transfer_queue: MutableSequence[tuple[Requisition, Exchange]] = deque(maxlen=16)
        ("")

    def clear(self) -> None:
        self.last_requisition = None
        self.last_exchange = None
        self.last_transmit = None
        self.last_transfer = None
        self.last_requisition_queue.clear()
        self.last_exchange_queue.clear()
        self.last_transmit_queue.clear()
        self.last_transfer_queue.clear()

    def _send(self, p: SendParams) -> Exchange:
        exchange = None
        try:
            exchange = super()._send(p)
        finally:
            self.last_requisition = p.requisition
            self.last_transmit = (p.requisition, exchange)
            self.last_requisition_queue.append(self.last_requisition)
            self.last_transmit_queue.append(self.last_transmit)
            if exchange is None:
                self.last_exchange = None
                self.last_transfer = None
            else:
                self.last_exchange = exchange
                self.last_transfer = (p.requisition, exchange)
                self.last_exchange_queue.append(self.last_exchange)
                self.last_transfer_queue.append(self.last_transfer)
        if exchange is None:
            raise Exception
        return exchange



class Last:
    @property
    def requisition(self) -> Optional[Requisition]:
        return self.recorder.last_requisition
    @property
    def exchange(self) -> Optional[Exchange]:
        return self.recorder.last_exchange
    @property
    def request(self) -> Optional[Request]:
        xchg = self.recorder.last_exchange
        if xchg is None:
            return None
        return xchg.request
    @property
    def response(self) -> Optional[Response]:
        xchg = self.recorder.last_exchange
        if xchg is None:
            return None
        return xchg.response
    @property
    def transmit(self) -> Optional[tuple[Requisition, Optional[Exchange]]]:
        return self.recorder.last_transmit
    @property
    def transfer(self) -> Optional[tuple[Requisition, Exchange]]:
        return self.recorder.last_transfer
    @property
    def requisition_queue(self) -> MutableSequence[Requisition]:
        return self.recorder.last_requisition_queue
    @property
    def exchange_queue(self) -> MutableSequence[Exchange]:
        return self.recorder.last_exchange_queue
    @property
    def transmit_queue(self) -> MutableSequence[tuple[Requisition, Optional[Exchange]]]:
        return self.recorder.last_transmit_queue
    @property
    def transfer_queue(self) -> MutableSequence[tuple[Requisition, Exchange]]:
        return self.recorder.last_transfer_queue

    def clear(self) -> None:
        self.recorder.clear()

    def __init__(self, recorder: Recorded) -> None:
        self.recorder: Recorded = recorder




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.user_SYNC import User
    from ...models.moderated_subreddit import ModeratedSubreddit

from .get_user_summary_SYNC import GetUserSummary
from .bulk_fetch_user_summary_SYNC import BulkFetchUserSummary
from .pull_SYNC import Pull
from .pull_subreddits_SYNC import PullSubreddits
from ...model_loaders.user_SYNC import load_user, load_potentially_suspended_user
from ...model_loaders.moderated_subreddit import load_moderated_subreddit
from ... import http
from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.user.sync1 import UserSearchPaginator
from ...exceptions import RejectedResultException

class UserProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.get_user_summary: GetUserSummary = GetUserSummary(client)
        ("""
            Get a partial user object by ID.

            .. .PARAMETERS

            :param `int` idn:

            .. .RETURNS

            :returns:
                * :class:`~.models.user_summary.UserSummary`
                * `None` if the user was not found.
            :rtype: `Optional`\\[:class:`~.models.user_summary.UserSummary`]
            """)
        self.bulk_fetch_user_summary: BulkFetchUserSummary = BulkFetchUserSummary(client)
        ("""
            Bulk fetch partial user objects, by ID.

            Any ID that can't be resolved will be ignored.
            Duplicate IDs in a batch will be ignored.

            .. .PARAMETERS

            :param `Iterable[int]` ids:

            .. .RETURNS

            :rtype: :class:`~.iterators.call_chunk_chaining_iterator.CallChunkChainingIterator`\\[:class:`~.models.user_summary.UserSummary`]
            """)
        self.pull: Pull = Pull(client)
        ("""
            Pull submissions and comments associated with a user.
            """)
        self.pull_subreddits: PullSubreddits = PullSubreddits(client)
        ("""
            Get user subreddits.
            """)

    def get_by_name(self, name: str) -> Optional[User]:
        """Get information about a user by name.

        Returns `None` if the specified user was not found,
        or is a suspended account.

        .. .PARAMETERS

        :param `str` name:
            User name.

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.user_SYNC.User`]
        """
        try:
            root = self._client.request('GET', f'/user/{name}/about')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        obj_data = root['data']
        if obj_data.get('is_suspended', False):
            return None
        return load_user(obj_data, self._client)

    def fetch_by_name(self, name: str) -> User:
        """Fetch information about a user by name.

        .. .PARAMETERS

        :param `str` name:
            User name.

        .. .RETURNS

        :rtype: :class:`~.models.user_SYNC.User`

        .. .RAISES

        :raises redditwarp.exceptions.RejectedResultException:
            The specified user account is suspended.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The specified user was not found.
        """
        root = self._client.request('GET', f'/user/{name}/about')
        obj_data = root['data']
        if obj_data.get('is_suspended', False):
            raise RejectedResultException('user is suspended')
        return load_user(obj_data, self._client)

    def get_potentially_suspended_by_name(self, name: str) -> Optional[object]:
        """Get information about a potentially suspended user, by name.

        Returns `None` if the specified user was not found.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :returns:
            * :class:`~.models.user_SYNC.User` if the user exists.
            * :class:`~.models.user_SYNC.SuspendedUser` if the user is suspended.
            * `None` if the user was not found.
        :rtype: `Optional`\\[`object`]
        """
        try:
            root = self._client.request('GET', f'/user/{name}/about')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        return load_potentially_suspended_user(root['data'], self._client)

    def fetch_potentially_suspended_by_name(self, name: str) -> object:
        """Fetch information about a potentially suspended user, by name.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :returns:
            * :class:`~.models.user_SYNC.User` if the user exists.
            * :class:`~.models.user_SYNC.SuspendedUser` if the user is suspended.
        :rtype: `object`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The specified user was not found.
        """
        root = self._client.request('GET', f'/user/{name}/about')
        return load_potentially_suspended_user(root['data'], self._client)

    def moderating(self, name: str) -> Sequence[ModeratedSubreddit]:
        """Get a list of partial subreddit objects the target user is a moderator of.

        This endpoint isn't very reliable on users with big lists.

        The specified user's own user subreddit won't be returned, but any other user
        subreddit they moderate will be.

        .. seealso::

           :meth:`p.account.pull_subreddits.moderating() <.pull_subreddits_SYNC.PullSubreddits.moderating>`
              For obtaining a list of the current user's moderated subreddits.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.moderated_subreddit.ModeratedSubreddit`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The specified user was not found.
        """
        root = self._client.request('GET', f'/user/{name}/moderated_subreddits')
        if not root:
            return ()
        return [load_moderated_subreddit(d) for d in root['data']]

    def search(self, query: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[UserSearchPaginator, User]:
        """Search users by name or description.

        .. .PARAMETERS

        :param `str` query:
        :param `Optional[int]` amount:

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`\\[:class:`~.pagination.paginators.user.sync1.UserSearchPaginator`, :class:`~.models.user_SYNC.User`]
        """
        p = UserSearchPaginator(self._client, '/users/search', query)
        return ImpartedPaginatorChainingIterator(p, amount)

    def exists(self, name: str) -> bool:
        """Check whether a user name exists.

        .. .PARAMETERS

        :param `str` name:

        .. .RETURNS

        :rtype: `bool`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `BAD_USERNAME`:
                * The specified parameter is empty.
                * The specified username contains illegal characters.
        """
        return not self._client.request('GET', '/api/username_available', params={'user': name})




from __future__ import annotations
from typing import Mapping, Any, Sequence, Optional

from datetime import datetime, timezone
from functools import cached_property

from ..core.const import AUTHORIZATION_BASE_URL

class CustomFeed:
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.owner: str = d['owner']
        ("""
            The curator of the custom feed.
            """)
        full_id36: str = d['owner_id']
        _, _, id36 = full_id36.partition('_')
        self.owner_id36: str = id36
        ("")
        self.owner_id: int = int(id36, 36)
        ("")
        self.name: str = d['name']
        ("")
        self.title: str = d['display_name']
        ("")
        self.description: str = d['description_md']
        ("")
        self.description_html: str = d['description_html']
        ("")
        self.subscriber_count: int = d['num_subscribers']
        ("")
        self.icon_url: str = d['icon_url']
        ("")
        self.subreddit_names: Sequence[str] = [o['name'] for o in d['subreddits']]
        ("")
        self.created_ut: int = int(d['created_utc'])
        ("")
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("")
        self.public: bool = d['visibility'] == 'public'
        ("")
        self.nsfw: bool = d['over_18']
        ("")
        self.path: str = d['path']
        ("")
        self.permalink: str = AUTHORIZATION_BASE_URL + d['path']
        ("")
        self.copied_from: Optional[tuple[str, str]] = None
        ("""
            A tuple of (user, feed) if the custom feed is a copy.
            """)
        copied_from_path = d['copied_from']
        if copied_from_path:
            user, feed = copied_from_path.strip('/').split('/')[1::2]
            self.copied_from = (user, feed)

    class Me:
        def __init__(self, outer: CustomFeed) -> None:
            d = outer.d
            self.can_edit: bool = d['can_edit']
            ("")
            self.favorited: bool = d['is_favorited']
            ("""
                Whether the current user has favourited the custom feed. False if no user context.
                """)
            self.subscribed: bool = d['is_subscriber']
            ("""
                Whether the current user is following the custom feed. False if no user context.
                """)

    me: cached_property[Me] = cached_property(Me)




from __future__ import annotations
from typing import Mapping, Any

from ..models.wiki import (
    WikiPageRevisionAuthorUser,
    WikiPage,
    WikiPageRevision,
    WikiPageSettings,
)


def load_wiki_page_revision_author_user(d: Mapping[str, Any]) -> WikiPageRevisionAuthorUser:
    return WikiPageRevisionAuthorUser(d)


def load_wiki_page(d: Mapping[str, Any]) -> WikiPage:
    return WikiPage(
        d=d,
        body=d['content_md'],
        body_html=d['content_html'],
        can_revise=d['may_revise'],
        revision_uuid=d['revision_id'],
        revision_unixtime=d['revision_date'],
        revision_author=load_wiki_page_revision_author_user(d['revision_by']['data']),
        revision_message=d['reason'] or '',
    )

def load_wiki_page_revision(d: Mapping[str, Any]) -> WikiPageRevision:
    return WikiPageRevision(
        d=d,
        uuid=d['id'],
        unixtime=d['timestamp'],
        author=load_wiki_page_revision_author_user(d['author']['data']),
        message=d['reason'] or '',
        hidden=d['revision_hidden'],
    )


def load_wiki_page_settings(d: Mapping[str, Any]) -> WikiPageSettings:
    return WikiPageSettings(
        permlevel=d['permlevel'],
        editors=[load_wiki_page_revision_author_user(m['data']) for m in d['editors']],
        indexed=d['listed'],
    )




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
if TYPE_CHECKING:
    from .requisition import Requisition
    from .request import Request
    from .response import Response

from dataclasses import dataclass


@dataclass(repr=False, eq=False)
class Exchange:
    requisition: Requisition
    request: Request
    response: Response
    history: Sequence[Response]




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Optional
if TYPE_CHECKING:
    from datetime import datetime as DateTime

from dataclasses import dataclass
from enum import Enum, auto

from .datamemento import DatamementoDataclassesMixin

class CommentMessageCause(Enum):
    USERNAME_MENTION = auto()
    SUBMISSION_REPLY = auto()
    COMMENT_REPLY = auto()

@dataclass(repr=False, eq=False)
class MailboxMessage(DatamementoDataclassesMixin):
    d: Mapping[str, Any]
    subject: str
    ("""
        Subject of the message.

        If a comment message (i.e., :class:`CommentMessage`) then the value will be:

        * `comment reply` if a comment reply.
        * `post reply` if a post reply.
        * `username mention` if a username mention.
        """)
    author_name: str
    ("""
        Author name.

        Unknown what happens if the user is deleted.
        Is the value `[deleted]`, an empty string, or does the field in the
        underlying object not exist?
        If you have any information about this, please open an issue report at
        `<https://github.com/Pyprohly/redditwarp/issues>`_.
        """)
    unread: bool

@dataclass(repr=False, eq=False)
class ComposedMessage(MailboxMessage):
    id: int
    unixtime: int
    ("""
        Unix timestamp of when the composed message was sent.
        """)
    datetime: DateTime
    ("""
        A datetime object of when the composed message was sent.
        """)
    body: str
    ("""
        The message content.
        """)
    body_html: str
    ("""
        The message content as HTML.
        """)
    unread: bool
    ("""
        Whether the message is new.
        """)
    distinguished: str
    ("""
        Either: `moderator`, `admin`, `gold-auto`, or empty string.

        Empty string if not distinguished.
        """)
    src_user_name: Optional[str]
    ("""
        Username of the sender.
        """)
    src_subr_name: Optional[str]
    ("""
        Name of the subreddit in which the message was sent from.
        """)
    dst_user_name: Optional[str]
    ("""
        Username of the recipient.
        """)
    dst_subr_name: Optional[str]
    ("""
        Name of the subreddit in which the message was sent to.
        """)
    src_user_id: Optional[int]
    ("""
        ID of the user who sent the message.
        """)

@dataclass(repr=False, eq=False)
class CommentMessage(MailboxMessage):
    @dataclass(repr=False, eq=False)
    class Submission:
        id: int
        title: str
        permalink_path: str
        permalink: str
        comment_count: int

    @dataclass(repr=False, eq=False)
    class Comment:
        id: int
        created_ut: int
        created_at: DateTime
        author_name: str
        ("""
            Author name.

            Unknown what happens if the user is deleted.
            Is the value `[deleted]`, an empty string, or does the field in the
            underlying object not exist?
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        author_id: int
        ("""
            Author ID.

            Unknown what happens if the user is deleted.
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        subreddit_name: str
        ("""
            Subreddit name.

            Unknown what happens if the subreddit is deleted.
            Is the value `[deleted]`, an empty string, or does the field in the
            underlying object not exist?
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        permalink_path: str
        permalink: str
        is_top_level: bool
        has_parent_comment: bool
        parent_comment_id36: str
        parent_comment_id: int
        score: int
        body: str
        body_html: str
        voted: int

    cause: CommentMessageCause
    submission: CommentMessage.Submission
    comment: CommentMessage.Comment




from __future__ import annotations
from typing import Mapping, Any, Optional

from datetime import datetime, timezone

from .datamemento import DatamementoPropertiesMixin

class Subreddit(DatamementoPropertiesMixin):
    class Me:
        class Flair:
            def __init__(self, d: Mapping[str, Any]) -> None:
                self.shown: bool = False if (x := d['user_sr_flair_enabled']) is None else x
                ("""Whether the current user has their flair enabled in the subreddit.

                    Value is `False` if the subreddit object comes from a search item.""")

                flair_text: Optional[str] = d['user_flair_text']
                flair_css_class: Optional[str] = d['user_flair_css_class']

                self.template_uuid: Optional[str] = x if (x := d['user_flair_template_id']) else None
                ("""
                    The current user's flair template UUID in the subreddit.

                    Value is `None` if no flair template is being used.
                    """)
                self.text_mode: str = d['user_flair_type']
                ("""
                    Either `text` or `richtext`.
                    """)
                self.text: str = flair_text or ''
                ("""
                    The flair text.

                    Check if the value is an empty string to tell if a flair is being used.
                    """)
                self.css_class: str = flair_css_class or ''
                ("""
                    The current user's flair CSS class.

                    When a flair template is being used, the value of this field will be
                    that of the CSS class designated by the template. If the flair template
                    does not specify a CSS class then the value will be an empty string.
                    """)
                self.bg_color: str = '' if (x := d['user_flair_background_color']) == 'transparent' else x
                ("""
                    A background color hex string. E.g., `#46d160`.

                    If a flair template is not being used then the value will be an empty string.

                    If a flair template is being used and the background color is unset then
                    the value is an empty string.
                    """)
                self.fg_color_scheme: str = d['user_flair_text_color'] or ''
                ("""
                    Either `dark` or `light`, or empty string.

                    Value is empty string if a flair template is not being used.
                    """)

                self.has_had_flair_assigned_before_in_subreddit: bool = flair_text is not None
                ("""
                    Because of quirks in the API, we can tell if the user has ever had flair assigned before in the subreddit.
                    """)
                self.has_had_flair_css_class_assigned_before_in_subreddit_when_no_flair_template_assigned: bool = flair_css_class is not None
                ("""
                    Because of quirks in the API, we can tell if the user has ever had a flair CSS class
                    assigned before in the subreddit, but this only works if a flair template is not being used.
                    """)

        def __init__(self, d: Mapping[str, Any]) -> None:
            self.favorited: bool = d['user_has_favorited']
            ("")
            self.is_banned: bool = d['user_is_banned']
            ("")
            self.is_approved_user: bool = d['user_is_contributor']
            ("")
            self.is_moderator: bool = d['user_is_moderator']
            ("""
                Whether the current user is a moderator of the subreddit.
                """)
            self.is_muted: bool = d['user_is_muted']
            ("")
            self.is_subscribed: bool = d['user_is_subscriber']
            ("")
            self.sr_theme_enabled: bool = d['user_sr_theme_enabled']
            ("""
                Whether the current user allows subreddit custom CSS.

                This is the "allow subreddits to show me custom themes" preference in old reddit.
                """)
            self.flair: Subreddit.Me.Flair = self.Flair(d)
            ("""
                Information about the current user's flair settings in the subreddit.
                """)

    class SubredditFlair:
        def __init__(self, d: Mapping[str, Any]) -> None:
            self.user_flair_enabled: bool = d['user_flair_enabled_in_sr']
            ("""
                Whether user flairs are enabled in the subreddit.

                In old Reddit this is the flair option that says "enable user flair in this subreddit".

                Caution: Value is false if object was retrieved from a search. This isn't the case with
                the other variables.
                """)
            self.user_flair_position: str = d['user_flair_position']
            ("""
                Either `left` or `right`, or empty string.

                Starts off as `right` in new subreddits.

                Can be set to an empty string via API calls
                (see :meth:`~.redditwarp.siteprocs.flair.SYNC.FlairProcedures.configure_subreddit_flair_settings`)
                but not through the UI.
                If an empty string then all user flairs are hidden, despite the :attr:`user_flair_enabled` setting.
                """)
            self.user_flair_self_assign: bool = d['can_assign_user_flair']
            ("""
                Whether or not users can assign a flair to themselves in this subreddit.

                If false, only a moderator can assign flairs to users.

                In old Reddit this is the flair option that says "allow users to assign their own flair".
                """)

            self.post_flair_enabled: bool = d['link_flair_enabled']
            ("""Whether post flairs are enabled in the subreddit.

                In old Reddit, this field is tied to the 'link flair position' flair setting:
                the value is false when set to `none`.
                """)
            self.post_flair_position: str = d['link_flair_position']
            ("""
                Either `left` or `right`, or empty string.

                Value is empty string if :attr:`post_flair_enabled` is false (the 'none' option in the old Reddit UI).
                """)
            self.post_flair_self_assign: bool = d['can_assign_link_flair']
            ("""
                Whether or not users can assign a flair to their submission in this subreddit.

                If false, only a moderator can assign flairs to submissions.

                In old Reddit this is the flair option that says "allow submitters to assign their own link flair".
                """)

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id36: str = d['id']
        ("""
            The ID of the subreddit as a base 36 number.
            """)
        self.id: int = int(self.id36, 36)
        ("""
            The subreddit ID.
            """)
        self.name: str = d['display_name']
        ("""
            The name of the subreddit.
            """)
        self.openness: str = d['subreddit_type']
        ("""
            Either: `public`, `private`, `restricted`, `archived`,
            `employees_only`, `gold_only`, `gold_restricted`, or `user`.
            """)
        self.created_ut: int = int(d['created_utc'])
        ("""
            Unix timestamp of when the subreddit was created.
            """)
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("""
            When the subreddit was created.
            """)

        self.title: str = d['title']
        ("""
            The title of the subreddit.

            Max. chars.: 100.
            """)
        self.public_description: str = d['public_description']
        ("")
        self.public_description_html: str = d['public_description_html']
        ("")
        self.sidebar_description: str = d['description']
        ("")
        self.sidebar_description_html: str = d['description_html']
        ("")
        self.submitting_form_note: str = d['submit_text']
        ("")
        self.submitting_form_note_html: str = d['submit_text_html']
        ("")
        self.submit_text_label: str = d['submit_text_label'] or ''
        ("""
            Custom label text for the "Submit a new text post" button.

            In old Reddit this is the "Custom label for submit text post button" subreddit option.
            """)
        self.submit_link_label: str = d['submit_link_label'] or ''
        ("""
            Custom label text for the "Custom label for submit link button" button.

            In old Reddit this is the "Custom label for submit text post button" subreddit option.
            """)

        self.subscriber_count: int = d['subscribers']
        ("")
        self.viewing_count: int = -1 if (x := d['active_user_count']) is None else x
        ("""
            The number of online users who are subscribed to the subreddit.

            Value is `-1` if object was retrieved from a search.
            """)

        submission_type: str = d['submission_type']
        self.accepts_text_submissions: bool = submission_type in ('any', 'self')
        ("")
        self.accepts_link_submissions: bool = submission_type in ('any', 'link')
        ("")
        self.accepts_gallery_submissions: bool = d['allow_galleries']
        ("")
        self.accepts_poll_submissions: bool = d['allow_polls']
        ("")

        self.suggested_comment_sort: str = d['suggested_comment_sort'] or ''
        ("""
            Either: `confidence` (best), `new`, `old`, `top`, `qa`, `controversial`, or `live`.

            Value is empty string if not set.
            """)

        self.nsfw: bool = d['over18']
        ("")
        self.quarantined: bool = d['quarantine']
        ("")

        self.me: Optional[Subreddit.Me] = None
        ("""
            Attributes related to the current user.
            """)
        if d['user_is_moderator'] is not None:
            self.me = self.Me(d)

        self.flair: Subreddit.SubredditFlair = self.SubredditFlair(d)
        ("")

        self.has_menu_widget: bool = d['has_menu_widget']
        ("""
            True if the subreddit has a menu widget active, otherwise false.

            Value is false if the object was retrieved from a search.
            """)

Community = Subreddit

class InaccessibleSubreddit(DatamementoPropertiesMixin):
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self.id36: str = d['id']
        ("")
        self.id: int = int(self.id36, 36)
        ("")
        self.created_ut: int = int(d['created_utc'])
        ("")
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        ("")

        self.name: str = d['display_name']
        ("")
        self.openness: str = d['subreddit_type']
        ("")

        self.title: str = d['title']
        ("")

InaccessibleCommunity = InaccessibleSubreddit




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, Optional, Mapping, Any
if TYPE_CHECKING:
    from ..client_SYNC import Client

from functools import cached_property

from .comment_tree_SYNC import MoreCommentsTreeNode

class MoreComments:
    @cached_property
    def submission_id(self) -> int:
        return int(self.submission_id36, 36)

    @cached_property
    def comment_id(self) -> int:
        return int(self.comment_id36, 36)

    def __init__(self,
        *,
        submission_id36: str,
        comment_id36: str,
        sort: str,
        d: Mapping[str, Any],
        client: Client,
    ):
        self.submission_id36: str = submission_id36
        ("")
        self.comment_id36: str = comment_id36
        ("")
        self.sort: str = sort
        ("")
        self.d: Mapping[str, Any] = d
        ("")
        self.client: Client = client
        ("")

    def __call__(self, *,
        depth: Optional[int] = None,
    ) -> MoreCommentsTreeNode:
        raise NotImplementedError

class ContinueThisThread(MoreComments):
    def __call__(self, *,
        depth: Optional[int] = None,
    ) -> MoreCommentsTreeNode:
        node = self.client.p.comment_tree.low.fetch(self.submission_id36, self.comment_id36)
        return MoreCommentsTreeNode(None, node.children[0].children, node.more)

class LoadMoreComments(MoreComments):
    def __init__(self,
        *,
        submission_id36: str,
        comment_id36: str,
        child_id36s: Sequence[str],
        sort: str,
        count: int,
        d: Mapping[str, Any],
        client: Client,
    ):
        super().__init__(
            submission_id36=submission_id36,
            comment_id36=comment_id36,
            sort=sort,
            d=d,
            client=client,
        )
        self.child_id36s: Sequence[str] = child_id36s
        ("")
        self.count: int = count
        ("")

    def __call__(self, *,
        depth: Optional[int] = None,
        exact: bool = False,
    ) -> MoreCommentsTreeNode:
        return self.client.p.comment_tree.low.more_children(
            self.submission_id36,
            self.child_id36s,
            sort=self.sort,
            depth=depth,
            exact=exact,
        )




from __future__ import annotations
from typing import Mapping, Any

from ..http.util.json_loading import load_json_from_response_but_prefer_status_code_exception_on_failure
from ..auth.token_obtainment_client_SYNC import TokenObtainmentClient
from ..auth.utils import apply_basic_auth
from .exceptions import raise_for_reddit_token_server_response_error, raise_for_reddit_auth_response_exception

class RedditTokenObtainmentClient(TokenObtainmentClient):
    def fetch_data(self) -> Mapping[str, Any]:
        headers: dict[str, str] = {}
        apply_basic_auth(headers, *self.client_creds)
        xchg = self.http.inquire('POST', self.url, headers=headers, data=self.grant)
        resp = xchg.response

        try:
            json_data = load_json_from_response_but_prefer_status_code_exception_on_failure(resp)
            raise_for_reddit_token_server_response_error(json_data)
            resp.ensure_successful_status()

        except Exception as cause:
            try:
                raise_for_reddit_auth_response_exception(cause, xchg)
            except Exception as exc:
                raise exc from cause
            raise

        return json_data




from __future__ import annotations
from typing import Any, Mapping, Collection

from functools import cached_property
from datetime import datetime, timezone

class ModeratorListItem:
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        full_id36: str = d['id']
        _, _, id36 = full_id36.partition('_')
        self.id36: str = id36
        ("""
            User ID of the moderator as a base 36 number.
            """)
        self.id: int = int(id36, 36)
        ("""
            User ID of the moderator.
            """)
        self.name: str = d['name']
        ("""
            The moderator's user name.
            """)
        #self.rel_id: str = d['rel_id']
        self.added_ut: int = int(d['date'])
        ("""
            Unix timestamp of when the user was added as a moderator.
            """)
        self.flair_text: str = d['author_flair_text']
        ("""
            The moderator's flair text.
            """)
        self.flair_css_class: str = d['author_flair_css_class']
        ("""
            The moderator's flair CSS class.
            """)
        self.permissions: Collection[str] = d['mod_permissions']
        ("""
            Values: `all`, `access`, `chat_config`, `chat_operator`, `config`, `flair`, `mail`, `posts`, `wiki`.
            """)

    @cached_property
    def added_at(self) -> datetime:
        """Datetime object of when the user was added as a moderator."""
        return datetime.fromtimestamp(self.added_ut, timezone.utc)




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional
if TYPE_CHECKING:
    from .more_comments_ASYNC import MoreComments

from dataclasses import dataclass

from ..util.tree_node import TreeNode
from .comment_ASYNC import Comment
from .submission_ASYNC import Submission

TValue_co = TypeVar('TValue_co', covariant=True)

@dataclass(repr=False, eq=False, frozen=True)
class CommentSubtreeTreeNode(TreeNode[TValue_co, 'CommentTreeNode']):
    more: Optional[MoreComments]

class MoreCommentsTreeNode(CommentSubtreeTreeNode[None]):
    pass

class CommentTreeNode(CommentSubtreeTreeNode[Comment]):
    pass

class SubmissionTreeNode(CommentSubtreeTreeNode[Submission]):
    pass




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
if TYPE_CHECKING:
    from .rules import Rule

from dataclasses import dataclass

from .widget_ASYNC import Widget

@dataclass(repr=False, eq=False)
class RulesWidget(Widget):
    display: str
    rules: Sequence[Rule]




from . import handler_SYNC  # noqa: F401
from .handler_SYNC import Handler as Handler  # noqa: F401
from . import delegating_handler_SYNC  # noqa: F401
from .delegating_handler_SYNC import DelegatingHandler as DelegatingHandler  # noqa: F401
from . import http_client_SYNC  # noqa: F401
from .http_client_SYNC import HTTPClient as HTTPClient  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.user_relationship import (
        UserRelationship,
        BannedSubredditUserRelationship,
    )

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.moderation.sync1 import (
    UserRelationshipListingPaginator,
    BannedSubredditUserRelationshipListingPaginator,
)

class LegacyPullUsers:
    def __init__(self, client: Client) -> None:
        self._client = client

    def approved_users(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[UserRelationshipListingPaginator, UserRelationship]:
        p = UserRelationshipListingPaginator(self._client, f'/r/{sr}/about/contributors')
        return ImpartedPaginatorChainingIterator(p, amount)

    def wiki_contributors(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[UserRelationshipListingPaginator, UserRelationship]:
        p = UserRelationshipListingPaginator(self._client, f'/r/{sr}/about/wikicontributors')
        return ImpartedPaginatorChainingIterator(p, amount)

    def banned(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[BannedSubredditUserRelationshipListingPaginator, BannedSubredditUserRelationship]:
        p = BannedSubredditUserRelationshipListingPaginator(self._client, f'/r/{sr}/about/banned')
        return ImpartedPaginatorChainingIterator(p, amount)

    def muted(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[UserRelationshipListingPaginator, UserRelationship]:
        p = UserRelationshipListingPaginator(self._client, f'/r/{sr}/about/muted')
        return ImpartedPaginatorChainingIterator(p, amount)

    def wiki_banned(self, sr: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[BannedSubredditUserRelationshipListingPaginator, BannedSubredditUserRelationship]:
        p = BannedSubredditUserRelationshipListingPaginator(self._client, f'/r/{sr}/about/wikibanned')
        return ImpartedPaginatorChainingIterator(p, amount)
